#!/usr/bin/env python
# -*- coding: utf-8 -*-

import numpy as np
from tqdm import tqdm
from loguru import logger
import torch
import torch.nn as nn
from torch.nn import CrossEntropyLoss, MSELoss, BCEWithLogitsLoss
from transformers import BertConfig, BertTokenizerFast
from transformers.modeling_bert import BertPreTrainedModel, BertModel
from sklearn.metrics import classification_report, roc_auc_score

from ..trainer import Trainer, get_default_optimizer_parameters
from ...losses import FocalLoss
from ...utils import softmax, sigmoid, acc_and_f1
from ...utils.multiprocesses import barrier_leader_process, barrier_member_processes


def logits_to_preds(logits):
    preds = np.argmax(logits, axis=1)
    probs = softmax(logits)
    # 调整类别0的判断阈值
    #  preds = np.array([
    #      0 if x == 1 and prob[1] < 0.60 else x for prob, x in zip(probs, preds)
    #  ])
    #  preds = np.array([
    #      1 if x == 0 and prob[1] < 0.90 else x for prob, x in zip(probs, preds)
    #  ])

    return preds


class BertForSequenceClassification(BertPreTrainedModel):
    def __init__(self, config):
        super().__init__(config)
        self.num_labels = config.num_labels

        self.bert = BertModel(config)
        self.dropout = nn.Dropout(config.hidden_dropout_prob)
        self.classifier = nn.Linear(config.hidden_size, config.num_labels)
        self.loss_type = config.loss_type
        self.focalloss_gamma = config.focalloss_gamma
        self.focalloss_alpha = config.focalloss_alpha

        self.init_weights()

    #  @add_start_docstrings_to_callable(BERT_INPUTS_DOCSTRING)
    def forward(
        self,
        input_ids=None,
        attention_mask=None,
        token_type_ids=None,
        position_ids=None,
        head_mask=None,
        inputs_embeds=None,
        labels=None,
    ):
        r"""
        labels (:obj:`torch.LongTensor` of shape :obj:`(batch_size,)`, `optional`, defaults to :obj:`None`):
            Labels for computing the sequence classification/regression loss.
            Indices should be in :obj:`[0, ..., config.num_labels - 1]`.
            If :obj:`config.num_labels == 1` a regression loss is computed (Mean-Square loss),
            If :obj:`config.num_labels > 1` a classification loss is computed (Cross-Entropy).

    Returns:
        :obj:`tuple(torch.FloatTensor)` comprising various elements depending on the configuration (:class:`~transformers.BertConfig`) and inputs:
        loss (:obj:`torch.FloatTensor` of shape :obj:`(1,)`, `optional`, returned when :obj:`label` is provided):
            Classification (or regression if config.num_labels==1) loss.
        logits (:obj:`torch.FloatTensor` of shape :obj:`(batch_size, config.num_labels)`):
            Classification (or regression if config.num_labels==1) scores (before SoftMax).
        hidden_states (:obj:`tuple(torch.FloatTensor)`, `optional`, returned when ``config.output_hidden_states=True``):
            Tuple of :obj:`torch.FloatTensor` (one for the output of the embeddings + one for the output of each layer)
            of shape :obj:`(batch_size, sequence_length, hidden_size)`.

            Hidden-states of the model at the output of each layer plus the initial embedding outputs.
        attentions (:obj:`tuple(torch.FloatTensor)`, `optional`, returned when ``config.output_attentions=True``):
            Tuple of :obj:`torch.FloatTensor` (one for each layer) of shape
            :obj:`(batch_size, num_heads, sequence_length, sequence_length)`.

            Attentions weights after the attention softmax, used to compute the weighted average in the self-attention
            heads.

    Examples::

        from transformers import BertTokenizer, BertForSequenceClassification
        import torch

        tokenizer = BertTokenizer.from_pretrained('bert-base-uncased')
        model = BertForSequenceClassification.from_pretrained('bert-base-uncased')

        input_ids = torch.tensor(tokenizer.encode("Hello, my dog is cute", add_special_tokens=True)).unsqueeze(0)  # Batch size 1
        labels = torch.tensor([1]).unsqueeze(0)  # Batch size 1
        outputs = model(input_ids, labels=labels)

        loss, logits = outputs[:2]

        """

        outputs = self.bert(
            input_ids,
            attention_mask=attention_mask,
            token_type_ids=token_type_ids,
            position_ids=position_ids,
            head_mask=head_mask,
            inputs_embeds=inputs_embeds,
        )

        pooled_output = outputs[1]

        pooled_output = self.dropout(pooled_output)
        logits = self.classifier(pooled_output)

        outputs = (logits, ) + outputs[
            2:]  # add hidden states and attention if they are here

        #  if labels is not None:
        if labels is not None and any(labels):
            if self.num_labels == 1:
                #  We are doing regression
                loss_fct = MSELoss()
                loss = loss_fct(logits.view(-1), labels.view(-1))
            else:

                #  #  logger.debug(f"logits: {type(logits)}, {logits}")
                #  preds = logits_to_preds(logits.detach().cpu().numpy())
                #  preds = torch.tensor(preds).float().cuda()
                #  loss = FocalLoss(gamma=1.5, alpha=None)(preds.view(-1),
                #                                          labels.view(-1))
                #

                #  loss_fct = CrossEntropyLoss()
                #  loss = loss_fct(logits.view(-1, self.num_labels),
                #                  labels.view(-1))

                #  logger.debug(f"preds: {type(preds)}, {preds}")

                assert self.loss_type in [
                    'FocalLoss', 'CrossEntropyLoss', 'BCEWithLogitsLoss'
                ]
                if self.loss_type == 'FocalLoss':
                    loss = FocalLoss(gamma=self.focalloss_gamma,
                                     alpha=self.focalloss_alpha)(logits.view(
                                         -1, self.num_labels), labels.view(-1))
                elif self.loss_type == 'BCEWithLogitsLoss':
                    loss_fct = BCEWithLogitsLoss()
                    #  logger.debug(f"logits: {logits.shape}, {logits}")
                    #  logger.debug(f"labels: {labels.shape}, {labels}")
                    loss = loss_fct(logits, labels.float())
                else:
                    loss_fct = CrossEntropyLoss()
                    loss = loss_fct(logits.view(-1, self.num_labels),
                                    labels.view(-1))

                #  from ...losses import FocalLoss
                #  #  loss = FocalLoss(gamma=1.5,
                #  #                   alpha=None)(logits.view(-1, self.num_labels),
                #  #                               labels.view(-1))
                #  loss = FocalLoss(
                #      gamma=1.5,
                #      alpha=[
                #          #          # --- 0.9354 (10/10)
                #          #          #  0.5,
                #          #          #  0.40,
                #          #          #  0.10,
                #          #          #  0.05,
                #          #          #  0.5,
                #          #          #  0.30,
                #          #          #  0.25,
                #          #          #  0.35
                #          #          # --- 0.9383 (10/10)
                #          #          0.5,
                #          #          0.40,
                #          #          0.10,
                #          #          0.05,
                #          #          0.5,
                #          #          0.40,
                #          #          0.40,
                #          #          0.40
                #          #  # focalloss_sympotom --- 0.9423 (10/10)
                #          #  # online f1: 0.92590
                #          #  0.50,  #   10
                #          #  0.40,  #   52
                #          #  0.10,  # 1267
                #          #  0.05,  # 2550
                #          #  0.50,  #    7
                #          #  0.40,  #  147
                #          #  0.35,  #  867
                #          #  0.40,  #  100
                #          #  # focalloss_sympotom --- 0.9396 (10/10)
                #          #  0.50,  #   10
                #          #  0.40,  #   52
                #          #  0.10,  # 1267
                #          #  0.05,  # 2550
                #          #  0.50,  #    7
                #          #  0.40,  #  147
                #          #  0.38,  #  867
                #          #  0.40,  #  100
                #          #  # focalloss_disease --- 0.9407 (7/10)
                #          #  # online f1: 0.92330
                #          #  0.50,  #   10
                #          #  0.40,  #   52
                #          #  0.08,  # 1267
                #          #  0.05,  # 2550
                #          #  0.50,  #    7
                #          #  0.40,  #  147
                #          #  0.35,  #  867
                #          #  0.40,  #  100
                #          #  # focalloss_others --- 0.9396 (10/10)
                #          #  0.50,  #   10
                #          #  0.40,  #   52
                #          #  0.10,  # 1267
                #          #  0.05,  # 2550
                #          #  0.50,  #    7
                #          #  0.40,  #  147
                #          #  0.35,  #  867
                #          #  0.30,  #  100
                #          # focalloss_checks --- 0.9397 (8/10)
                #          #  # online f1: 0.92640
                #          0.50,  #   10
                #          0.40,  #   52
                #          0.10,  # 1267
                #          0.05,  # 2550
                #          0.50,  #    7
                #          0.38,  #  147
                #          0.35,  #  867
                #          0.40,  #  100
                #      ])(logits.view(-1, self.num_labels), labels.view(-1))


#

#  glue_labels = ['病毒', '细菌', '疾病', '药物', '医学专科', '检查科目', '症状', 'NoneType']
# 10,     52,    1267,  2550,  7,     147,   867,   100
# [0.200, 0.200, 0.025, 0.025, 0.200, 0.100, 0.050, 0.200]

#  from ...losses import CircleLoss
#  loss = CircleLoss(m=0.25,
#                    gamma=128)
#  loss = CircleLoss(scale=32, margin=0.25)(logits.view(
#      -1, self.num_labels), labels.view(-1))
#  outputs = (loss, ) + outputs
            return (loss, ) + outputs
        else:
            # (loss), logits, (hidden_states), (attentions)
            #  return (torch.tensor(0.0).cuda(), ) + outputs
            return outputs

MODEL_CLASSES = {
    'bert': (BertConfig, BertForSequenceClassification, BertTokenizerFast),
    #  'albert':
    #  (AlbertConfig, AlBertForSequenceClassification, BertTokenizerFast),
}


def load_pretrained_tokenizer(args):
    config_class, model_class, tokenizer_class = MODEL_CLASSES[args.model_type]
    tokenizer = tokenizer_class.from_pretrained(
        args.model_path,
        do_lower_case=args.do_lower_case,
        cache_dir=args.cache_dir if args.cache_dir else None,
    )

    return tokenizer


def load_pretrained_model(args):
    # make sure only the first process in distributed training
    # will download model & vocab
    barrier_member_processes(args)

    config_class, model_class, tokenizer_class = MODEL_CLASSES[args.model_type]
    config = config_class.from_pretrained(
        args.model_path,
        num_labels=args.num_labels,
        cache_dir=args.cache_dir if args.cache_dir else None,
    )
    setattr(config, 'loss_type', args.loss_type)
    setattr(config, 'focalloss_gamma', args.focalloss_gamma)
    setattr(config, 'focalloss_alpha', args.focalloss_alpha)

    logger.info(f"model_path: {args.model_path}")
    logger.info(f"config:{config}")
    model = model_class.from_pretrained(
        args.model_path,
        from_tf=bool(".ckpt" in args.model_path),
        config=config,
        cache_dir=args.cache_dir if args.cache_dir else None)

    # make sure only the first process in distributed training
    # will download model & vocab
    barrier_leader_process(args)

    return model


def load_model(args):
    model = load_pretrained_model(args)
    model.to(args.device)
    return model


def build_default_model(args):
    """
    自定义模型
    规格要求返回模型(model)、优化器(optimizer)、调度器(scheduler)三元组。
    """

    # -------- model --------
    model = load_pretrained_model(args)
    model.to(args.device)

    # -------- optimizer --------
    from transformers.optimization import AdamW
    optimizer_parameters = get_default_optimizer_parameters(
        model, args.weight_decay)
    optimizer = AdamW(optimizer_parameters,
                      lr=args.learning_rate,
                      correct_bias=False)

    # -------- scheduler --------
    from transformers.optimization import get_linear_schedule_with_warmup

    scheduler = get_linear_schedule_with_warmup(
        optimizer,
        num_warmup_steps=args.total_steps * args.warmup_rate,
        num_training_steps=args.total_steps)

    return model, optimizer, scheduler


def collate_fn(batch):
    """
    batch should be a list of (sequence, target, length) tuples...
    Returns a padded tensor of sequences sorted from longest to shortest,
    """
    all_input_ids, all_attention_mask, all_token_type_ids, all_lens, all_labels = map(
        torch.stack, zip(*batch))
    max_len = max(all_lens).item()
    all_input_ids = all_input_ids[:, :max_len]
    all_attention_mask = all_attention_mask[:, :max_len]
    all_token_type_ids = all_token_type_ids[:, :max_len]
    #  all_labels = all_labels[:, :max_len]
    return all_input_ids, all_attention_mask, all_token_type_ids, all_labels


class ClassReporter(object):
    def __init__(self, target_names=None, labels=None):
        logger.debug(f"target_names: {target_names}")
        logger.debug(f"labels: {labels}")
        self.target_names = target_names
        self.labels = labels

    def __call__(self, output, target):
        #  _, y_pred = torch.max(output.data, 1)
        #  y_pred = y_pred.cpu().numpy()
        #  y_true = target.cpu().numpy()

        #  y_pred = output.cpu().numpy()
        #  y_true = target.cpu().numpy()

        y_pred = output
        y_true = target

        #  print(y_pred, y_true, self.labels, self.target_names)
        #  if len(y_true.shape) == 1:
        classify_report = classification_report(y_true,
                                                y_pred,
                                                labels=self.labels,
                                                target_names=self.target_names)
        logger.info(f"classify_report:\n{classify_report}")
        #  else:
        #      for i, label in enumerate(self.labels):
        #          auc = roc_auc_score(y_score=y_pred[:, i], y_true=y_true[:, i])
        #          print(f"{label} - auc: {auc:.4f}")

        return classify_report


def init_labels(args, labels):
    args.id2label = {i: label for i, label in enumerate(labels)}
    args.label2id = {label: i for i, label in enumerate(labels)}
    args.num_labels = len(args.label2id)

    logger.info(f"args.label2id: {args.label2id}")
    logger.info(f"args.id2label: {args.id2label}")
    logger.info(f"args.num_labels: {args.num_labels}")


class GlueTrainer(Trainer):
    def __init__(self, args, glue_labels, build_model=None, tokenizer=None):
        super(GlueTrainer, self).__init__(args)
        init_labels(args, glue_labels)

        if tokenizer:
            self.tokenizer = tokenizer
        else:
            self.tokenizer = load_pretrained_tokenizer(args)

        if build_model is None:
            self.build_model = build_default_model
        else:
            self.build_model = build_model

        self.label2id = args.label2id
        self.collate_fn = collate_fn

        self.logits = None
        self.preds = None
        self.probs = None

        self.class_reporter = ClassReporter(
            target_names=[x for x in self.label2id.keys()],
            labels=[x for x in self.label2id.values()])

    #  def generate_dataloader(self, args, dataset, batch_size, keep_order=True):
    #
    #      Sampler = SequentialSampler if keep_order else RandomSampler
    #      sampler = DistributedSampler(dataset) if is_multi_processes(
    #          args) else Sampler(dataset)
    #      dataloader = DataLoader(dataset,
    #                              sampler=sampler,
    #                              batch_size=batch_size,
    #                              collate_fn=collate_fn)
    #      return dataloader

    def examples_to_dataset(self, examples, max_seq_length):
        from .dataset import examples_to_dataset
        return examples_to_dataset(examples, self.label2id, self.tokenizer,
                                   max_seq_length)

    def batch_to_inputs(self, args, batch, known_labels=True):
        inputs = {
            "input_ids": batch[0],
            "attention_mask": batch[1],
            "labels": batch[3]
        }
        if args.model_type != "distilbert":
            # XLM, DistilBERT, RoBERTa, and XLM-RoBERTa don't use segment_ids
            inputs["token_type_ids"] = (batch[2] if args.model_type in [
                "bert", "xlnet", "albert"
            ] else None)
        return inputs

    def on_eval_start(self, args, eval_dataset):
        self.logits = None
        self.preds = None
        self.out_label_ids = None
        self.results = {}
        self.probs = None

    #  def on_eval_step(self, args, eval_dataset, step, model, inputs, outputs):
    def on_eval_step(self, args, model, step, batch, batch_features):
        input_ids, attention_mask, _, label_ids = batch
        outputs = self.on_train_step(args, model, step, batch)
        eval_loss, logits = outputs[:2]
        if self.logits is None:
            self.logits = logits.detach().cpu().numpy()
            self.out_label_ids = label_ids.detach().cpu().numpy()
        else:
            self.logits = np.append(self.logits,
                                    logits.detach().cpu().numpy(),
                                    axis=0)
            self.out_label_ids = np.append(self.out_label_ids,
                                           label_ids.detach().cpu().numpy(),
                                           axis=0)

        self.preds = np.argmax(self.logits, axis=1)
        self.probs = softmax(self.logits)

        # 调整类别0的判断阈值
        #  self.preds = np.array([
        #      0 if x == 1 and prob[1] < 0.60 else x
        #      for prob, x in zip(self.probs, self.preds)
        #  ])
        self.preds = logits_to_preds(self.logits)
        #  logger.debug(f"self.preds: {self.preds.shape}, {self.preds}")
        #  logger.debug(
        #      f"self.out_label_ids: {self.out_label_ids.shape}, {self.out_label_ids}, {np.argmax(self.out_label_ids, axis=1)}"
        #  )

        if len(self.out_label_ids.shape) == 1:
            result = acc_and_f1(self.preds, self.out_label_ids)
        else:
            result = acc_and_f1(self.preds,
                                np.argmax(self.out_label_ids, axis=1))

        self.results.update(result)

        return (eval_loss, ), self.results

    def on_eval_end(self, args, eval_dataset):
        if len(self.out_label_ids.shape) == 1:
            self.preds = np.argmax(self.logits, axis=1)
        else:
            self.preds = np.array([x > 0.5
                                   for x in sigmoid(self.logits)]).astype(int)
        # for regessions
        #  self.preds = np.squeeze(self.preds)

        #  logger.debug(f"self.preds: {self.preds.shape}, {self.preds}")
        #  logger.debug(
        #      f"self.out_label_ids: {self.out_label_ids.shape}, {self.out_label_ids}"
        #  )
        self.class_reporter(self.preds, self.out_label_ids)

        logger.info(f"  Num examples = {len(eval_dataset)}")
        logger.info(f"  Batch size = {args.eval_batch_size}")
        logger.info(f"******** Eval results ********")
        for key in self.results.keys():
            logger.info(f" dev: {key} = {self.results[key]:.4f}")

        return self.results

    def on_predict_start(self, args, test_dataset):
        self.logits = None

        self.pred_results = None
        self.probs = None

    #  def on_predict_step(self, args, test_dataset, step, model, inputs,
    #                      outputs):
    def on_predict_step(self, args, model, step, batch):
        inputs = self.batch_to_inputs(args, batch, known_labels=False)
        outputs = model(**inputs)

        #  logger.debug(f"inputs: {inputs}")
        #  logger.debug(f"outputs: {outputs}")
        logits = outputs[0]
        if self.logits is None:
            self.logits = logits.detach().cpu().numpy()
        else:
            self.logits = np.append(self.logits,
                                    logits.detach().cpu().numpy(),
                                    axis=0)

    def on_predict_end(self, args, test_dataset):
        self.pred_results = np.argmax(self.logits, axis=1)
        self.probs = softmax(self.logits)

        #  self.pred_results = np.array([
        #      0 if x == 1 and prob[1] < 0.60 else x
        #      for prob, x in zip(self.probs, self.pred_results)
        #  ])
        self.pred_results = logits_to_preds(self.logits)

        logger.debug(f"pred_results: {self.pred_results}")
        logger.debug(f"probs: {self.probs}")



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from .focalloss import FocalLoss
from .circleloss import CircleLoss
from .label_smoothing import LabelSmoothingCrossEntropy
from .dice_loss import DiceLoss



# -*- coding: utf-8 -*-
import requests

def easyai_stop_machine(machine_id, machine_key):
    url = 'https://www.easyaiforum.cn/api/control/stop_machine'
    data = {"machine_id": machine_id, "machine_key": machine_key}
    req = requests.post(url, data=data)
    res = req.json()
    if res['code'] == 1000:
        return True
    else:
        return False


def stop_machine():
    machine_id = '3547'
    machine_key = '6e69db81acc5e3dea49c7c1e6f2b87b5'
    easyai_stop_machine(machine_id, machine_key)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from pathlib import Path
import json
from loguru import logger
import numpy as np
import pandas as pd
import torch
from torch.utils.data import DataLoader, TensorDataset
from torch.utils.data import RandomSampler, SequentialSampler
from ...utils.multiprocesses import barrier_leader_process, barrier_member_processes, is_master_process, is_multi_processes


#  @dataclass(frozen=False)
class InputExample:
    """
    A single training/test example for simple sequence classification.

    Args:
        guid: Unique id for the example.
        text_a: string. The untokenized text of the first sequence. For single
            sequence tasks, only this sequence must be specified.
        text_b: (Optional) string. The untokenized text of the second sequence.
            Only must be specified for sequence pair tasks.
        label: (Optional) string. The label of the example. This should be
            specified for train and dev examples, but not for test examples.
    """

    #  guid: str
    #  text_a: str
    #  text_b: Optional[str] = None
    #  label: Optional[str] = None
    def __init__(self, guid, text_a, text_b=None, label=None):
        self.guid = guid
        self.text_a = text_a
        self.text_b = text_b
        self.label = label

    def to_json_string(self):
        """Serializes this instance to a JSON string."""
        return json.dumps(dataclasses.asdict(self), indent=2,
                          sort_keys=True) + "\n"


class InputFeatures(object):
    def __init__(self, input_ids_a, attention_mask_a, input_ids_b,
                 attention_mask_b, label_id):
        self.input_ids_a = input_ids_a
        self.attention_mask_a = attention_mask_a
        self.input_ids_b = input_ids_b
        self.attention_mask_b = attention_mask_b
        self.label_id = label_id

    def __repr__(self):
        return str(self.to_json_string())

    def to_dict(self):
        """Serializes this instance to a Python dictionary."""
        output = copy.deepcopy(self.__dict__)
        return output

    def to_json_string(self):
        """Serializes this instance to a JSON string."""
        return json.dumps(self.to_dict(), indent=2, sort_keys=True) + "\n"


def convert_examples_to_features(
    examples,
    #  glue_labels,
    label2id,
    tokenizer,
    max_length=512,
    task=None,
    output_mode=None,
    pad_on_left=False,
    pad_token=0,
    pad_token_segment_id=0,
    mask_padding_with_zero=True,
):
    """
    Loads a data file into a list of ``InputFeatures``

    Args:
        examples: List of ``InputExamples`` or ``tf.data.Dataset`` containing the examples.
        tokenizer: Instance of a tokenizer that will tokenize the examples
        max_length: Maximum example length
        task: GLUE task
        glue_labels: List of labels. Can be obtained from the processor using the ``processor.get_labels()`` method
        output_mode: String indicating the output mode. Either ``regression`` or ``classification``
        pad_on_left: If set to ``True``, the examples will be padded on the left rather than on the right (default)
        pad_token: Padding token
        pad_token_segment_id: The segment ID for the padding token (It is usually 0, but can vary such as for XLNet where it is 4)
        mask_padding_with_zero: If set to ``True``, the attention mask will be filled by ``1`` for actual values
            and by ``0`` for padded values. If set to ``False``, inverts it (``1`` for padded values, ``0`` for
            actual values)

    Returns:
        If the ``examples`` input is a ``tf.data.Dataset``, will return a ``tf.data.Dataset``
        containing the task-specific features. If the input is a list of ``InputExamples``, will return
        a list of task-specific ``InputFeatures`` which can be fed to the model.

    """
    #  label2id = {label: i for i, label in enumerate(glue_labels)}

    features = []
    for (ex_index, example) in enumerate(examples):
        len_examples = len(examples)
        if ex_index % 10000 == 0:
            logger.info("Writing example %d/%d" % (ex_index, len_examples))

        def example_to_feature(text_a, text_b):
            inputs = tokenizer.encode_plus(
                text_a,
                text_b,
                add_special_tokens=True,
                max_length=max_length,
                return_token_type_ids=True,
            )
            input_ids, token_type_ids = inputs["input_ids"], inputs[
                "token_type_ids"]

            # The mask has 1 for real tokens and 0 for padding tokens. Only real
            # tokens are attended to.
            attention_mask = [1 if mask_padding_with_zero else 0
                              ] * len(input_ids)

            # Zero-pad up to the sequence length.
            padding_length = max_length - len(input_ids)
            if pad_on_left:
                input_ids = ([pad_token] * padding_length) + input_ids
                attention_mask = ([0 if mask_padding_with_zero else 1] *
                                  padding_length) + attention_mask
                token_type_ids = ([pad_token_segment_id] *
                                  padding_length) + token_type_ids
            else:
                input_ids = input_ids + ([pad_token] * padding_length)
                attention_mask = attention_mask + (
                    [0 if mask_padding_with_zero else 1] * padding_length)
                token_type_ids = token_type_ids + ([pad_token_segment_id] *
                                                   padding_length)

            assert len(
                input_ids
            ) == max_length, "Error with input length {} vs {}".format(
                len(input_ids), max_length)
            assert len(
                attention_mask
            ) == max_length, "Error with input length {} vs {}".format(
                len(attention_mask), max_length)
            assert len(
                token_type_ids
            ) == max_length, "Error with input length {} vs {}".format(
                len(token_type_ids), max_length)

            return input_ids, attention_mask, token_type_ids

        input_ids_a, attention_mask_a, token_type_ids_a = example_to_feature(
            example.text_a, None)
        input_ids_b, attention_mask_b, token_type_ids_b = example_to_feature(
            example.text_b, None)

        label = 0
        if example.label is not None:
            if output_mode == "classification":
                label = label2id[example.label]
            elif output_mode == "regression":
                label = float(example.label)
            else:
                raise KeyError(output_mode)

        if ex_index < 3:
            logger.info("*** Example ***")
            logger.info("guid: %s" % (example.guid))

            tokens_a = tokenizer.tokenize(example.text_a)
            logger.info(f"tokens_a: {' '.join([str(x) for x in tokens_a])}")

            logger.info("input_ids_a: %s" %
                        " ".join([str(x) for x in input_ids_a]))
            logger.info("attention_mask_a: %s" %
                        " ".join([str(x) for x in attention_mask_a]))
            logger.info("token_type_ids_a: %s" %
                        " ".join([str(x) for x in token_type_ids_a]))

            tokens_b = tokenizer.tokenize(example.text_b)
            logger.info(f"tokens_b: {' '.join([str(x) for x in tokens_b])}")
            logger.info("input_ids_b: %s" %
                        " ".join([str(x) for x in input_ids_b]))
            logger.info("attention_mask_b: %s" %
                        " ".join([str(x) for x in attention_mask_b]))
            logger.info("token_type_ids_b: %s" %
                        " ".join([str(x) for x in token_type_ids_b]))

            logger.info("label: %s (id = %d)" % (example.label, label))

        features.append(
            InputFeatures(input_ids_a, attention_mask_a, input_ids_b,
                          attention_mask_b, label))

    return features


def examples_to_dataset(args,
                        examples,
                        glue_labels,
                        tokenizer,
                        max_seq_length,
                        evaluate=False,
                        output_mode="classification",
                        alias=""):
    # Make sure only the first process in distributed training process the dataset,
    # and the others will use the cache
    if not evaluate:
        barrier_member_processes(args)

    args.id2label = {i: label for i, label in enumerate(glue_labels)}
    args.label2id = {label: i for i, label in enumerate(glue_labels)}
    args.num_labels = len(glue_labels)

    # Load data features from cache or dataset file
    cached_features_file = Path(
        args.data_dir
    ) / f"cached_glue_{args.dataset_name}{'_' + alias if alias else ''}"
    if args.cache_features and cached_features_file.exists(
    ) and not args.overwrite_cache:
        logger.info(
            f"Loading features from cached file {cached_features_file}")
        features = torch.load(cached_features_file)
    else:
        logger.info(f"Creating features from dataset file at {args.data_dir}")

        features = convert_examples_to_features(
            examples,
            args.label2id,
            tokenizer,
            max_length=max_seq_length,
            output_mode=output_mode,
            pad_on_left=bool(
                args.model_type in ["xlnet"]),  # pad on the left for xlnet
            pad_token=tokenizer.pad_token_id,
            pad_token_segment_id=tokenizer.pad_token_type_id,
        )

        if args.cache_features and is_master_process(args):
            logger.info(
                f"Saving features into cached file {cached_features_file}")
            torch.save(features, cached_features_file)

    logger.info(
        f"convert {len(examples)} examples to {len(features)} features.")
    # Make sure only the first process in distributed training process the dataset,
    # and the others will use the cache
    if not evaluate:
        barrier_leader_process(args)

    # Convert to Tensors and build dataset
    all_input_ids_a = torch.tensor([f.input_ids_a for f in features],
                                   dtype=torch.long)
    all_attention_mask_a = torch.tensor([f.attention_mask_a for f in features],
                                        dtype=torch.long)
    all_input_ids_b = torch.tensor([f.input_ids_b for f in features],
                                   dtype=torch.long)
    all_attention_mask_b = torch.tensor([f.attention_mask_b for f in features],
                                        dtype=torch.long)
    if output_mode == "classification":
        all_label_ids = torch.tensor([f.label_id for f in features],
                                     dtype=torch.long)
    elif output_mode == "regression":
        all_label_ids = torch.tensor([f.label_id for f in features],
                                     dtype=torch.float)

    dataset = TensorDataset(all_input_ids_a, all_attention_mask_a,
                            all_input_ids_b, all_attention_mask_b,
                            all_label_ids)
    return dataset


def build_encode_dataset(args, examples, tokenizer, max_seq_length):

    all_input_ids = []
    all_attention_mask = []
    for example in examples:
        inputs = tokenizer.encode_plus(
            example.text_a,
            example.text_b,
            add_special_tokens=True,
            max_length=max_seq_length,
            return_token_type_ids=True,
        )
        input_ids, token_type_ids = inputs["input_ids"], inputs[
            "token_type_ids"]
        attention_mask = [1] * len(input_ids)

        # Zero-pad up to the sequence length.
        pad_token = tokenizer.pad_token_id
        pad_token_segment_id = tokenizer.pad_token_type_id,
        padding_length = max_seq_length - len(input_ids)
        input_ids = input_ids + ([pad_token] * padding_length)
        attention_mask = attention_mask + ([0] * padding_length)
        token_type_ids = token_type_ids + ([pad_token_segment_id] *
                                           padding_length)
        all_input_ids.append(input_ids)
        all_attention_mask.append(attention_mask)

    # Convert to Tensors and build dataset
    all_input_ids = torch.tensor(all_input_ids, dtype=torch.long)
    all_attention_mask = torch.tensor(all_attention_mask, dtype=torch.long)
    dataset = TensorDataset(all_input_ids, all_attention_mask)
    return dataset



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from .trainer import GlueTrainer, load_pretrained_tokenizer, load_pretrained_model, load_model, logits_to_preds
from .dataset import InputExample, examples_to_dataset 
from .args import get_args



# -*- coding: utf-8 -*-

#%% ------------------------------------------------------------
import re
import pandas as pd
import numpy as np
from collections import Counter
from tqdm import tqdm

#%% ------------------------------------------------------------
results_dir = "./results"
ensemble_results_file = f"{results_dir}/ensemble_results.csv"

#%% ------------------------------------------------------------
pd_results = pd.DataFrame()
#  for i in range(total_folds):
import glob
fold_results_files = glob.glob(f"{results_dir}/result_fold*.txt")
fold_results_files = sorted(fold_results_files)
total_folds = len(fold_results_files)
for i, results_file in enumerate(fold_results_files):
    print(f"{i}: {results_file}")
    pd_result = pd.read_csv(results_file,
                            names=['ID', 'PREDICTION'],
                            header=None)
    if i == 0:
        pd_results['ID'] = pd_result['ID']
    pd_results[f"P{i}"] = pd_result['PREDICTION']

#%% ------------------------------------------------------------
ensemble_results = []

for i, row in tqdm(pd_results.iterrows()):

    results = [
        re.sub("[^\w\*\-\+\.\(\)\（\）]", "", x)
        if type(x) == str and len(x) < 30 else ""
        for x in [row[f"P{i}"] for i in range(total_folds)]
    ]

    results = [x for x in results if len(x) > 0]
    num_results = len(results)
    if num_results == 0:
        ensemble_results.append([row['ID'], "NaN"])
        continue

    c = Counter(results)
    if len(c) == 1:
        ensemble_results.append([row['ID'], results[0]])
        #  ensemble_results.append([row['ID'], "NaN"])
        continue

    guess_results = [row['ID']]

    #  # ----------------
    # Best 0.821840830144845
    # 选项多于4个的，删除只有1张投票的选项。
    predictions = c.most_common()
    for m in predictions:
        if m[0] == "NaN":
            continue
        if len(m[0]) == 0:
            continue
        if len(predictions) < 4:
            guess_results.append(m[0])
        else:
            if m[1] > 1:
                guess_results.append(m[0])

    ensemble_results.append(guess_results)

#%% ------------------------------------------------------------
with open(ensemble_results_file, 'w') as F:
    for predictions in ensemble_results:
        if len(predictions) <= 1:
            line = f"\"{predictions[0]}\",\"NaN"
        else:
            line = f"\"{predictions[0]}\",\""
            for p in predictions[1:]:
                if type(p) == str and len(p) == 0:
                    break
                line += f"{p},"
            line = line[:-1]
        F.write(line + "\"\n")



#!/usr/bin/env python
# -*- coding: utf-8 -*-


class Config:

    #  def update(cfg):
    #      for a in cfg.__dict__:
    #          if not a.startswith('_'):
    #              cfg[a]

    SEED = 8864
    debug_extract_labels = False

    surfix = ""

    experiment = "resumes"
    data_dir = f"../user_data/seqlabel_datasets/{experiment}"
    results_dir = f"../prediction_result/{experiment}_result/seqlabel"
    output_dir = f"../user_data/{experiment}_output/seqlabel"
    train_data_json_file = f"../data/Dataset/train_data/train_data.json"

    category = ""
    top_k = 20
    s_threshold = 0.1
    e_threshold = 0.1  #0.1
    min_label_len = 8
    seg_len = 510
    seg_backoff = 400
    use_additional_section = True
    strictly_equal = False
    num_augments = 0
    surfix = f"_aug{num_augments}"

    # -------------------- seqlabel model train & test begin. ---------------
    total_folds = 5

    # 是否应该取训练数据最大长度或最大长度95%（99%）？
    max_seq_length = seg_len  # 510  #160

    # -------------------- seqlabel model train & test end. ---------------

    BERT_CHINESE_DIR = "../data/External/pretrained/tensorflow/bert-base-chinese"
    epochs = 6  # 训练轮次
    batch_size = 8  # 按GPU显存大小调整，建议8G 取16, 12G 取24，16G 取32

    learning_rate = 5e-5
    min_learning_rate = 1e-5
    #  learning_rate = 1e-5
    #  min_learning_rate = 1e-6

    # 反序文本和标签，训练新模型
    reverse_text = False
    # ls -l /dev/nvidia? | wc -l
    gpus = 1

    # TPU
    steps_per_update = 8  # 权重更新前梯度累积的步数。
    #  learning_rate = 5e-4  # 运用梯度累积时，适当调高学习率

    config_path = f"{BERT_CHINESE_DIR}/bert_config.json"
    checkpoint_path = f"{BERT_CHINESE_DIR}/bert_model.ckpt"
    dict_path = f"{BERT_CHINESE_DIR}/vocab.txt"


class Config_WORKCONTENT(Config):
    category = "WORKCONTENT"
    top_k = 20
    s_threshold = 0.1
    e_threshold = 0.01
    min_label_len = 8
    seg_len = 510
    seg_backoff = 400
    strictly_equal = False
    use_additional_section = True
    num_augments = 0
    surfix = f"_aug{num_augments}"


class Config_PROJRESP(Config):
    category = "PROJRESP"
    top_k = 20
    s_threshold = 0.1
    e_threshold = 0.1  #0.1
    min_label_len = 8
    seg_len = 510
    seg_backoff = 400
    use_additional_section = True
    strictly_equal = False
    num_augments = 0
    surfix = f"_aug{num_augments}"


def get_best_model_file_name(selected_fold: int) -> str:
    return f"{Config.output_dir}/best_model_fold{selected_fold}{surfix}.model"


def get_best_model_weights_file_name(selected_fold: int, category) -> str:
    return f"{Config.output_dir}/best_model_{category}_fold{selected_fold}{Config.surfix}.weights"


def get_result_file_name(selected_fold: int) -> str:
    return f"{Config.results_dir}/result_{Config.category}_fold{selected_fold}{Config.surfix}.txt"


def get_wrong_eval_file_name(selected_fold: int) -> str:
    return f"{Config.output_dir}/wrong_eval_fold{selected_fold}{Config.surfix}.txt"



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from typing import Tuple
import torch
import torch.nn as nn
from torch import nn, Tensor
import torch.nn.functional as F


class CircleLoss(nn.Module):
    def __init__(self, scale=32, margin=0.25, similarity='cos', **kwargs):
        super(CircleLoss, self).__init__()
        self.scale = scale
        self.margin = margin
        self.similarity = similarity

    def forward(self, feats, labels):
        assert feats.size(0) == labels.size(0), \
            f"feats.size(0): {feats.size(0)} is not equal to labels.size(0): {labels.size(0)}"

        m = labels.size(0)
        mask = labels.expand(m, m).t().eq(labels.expand(m, m)).float()
        pos_mask = mask.triu(diagonal=1)
        neg_mask = (mask - 1).abs_().triu(diagonal=1)
        if self.similarity == 'dot':
            sim_mat = torch.matmul(feats, torch.t(feats))
        elif self.similarity == 'cos':
            feats = F.normalize(feats)
            sim_mat = feats.mm(feats.t())
        else:
            raise ValueError('This similarity is not implemented.')

        pos_pair_ = sim_mat[pos_mask == 1]
        neg_pair_ = sim_mat[neg_mask == 1]

        alpha_p = torch.relu(-pos_pair_ + 1 + self.margin)
        alpha_n = torch.relu(neg_pair_ + self.margin)
        margin_p = 1 - self.margin
        margin_n = self.margin
        loss_p = torch.sum(torch.exp(-self.scale * alpha_p * (pos_pair_ - margin_p)))
        loss_n = torch.sum(torch.exp(self.scale * alpha_n * (neg_pair_ - margin_n)))
        loss = torch.log(1 + loss_p * loss_n)
        return loss


#  def convert_label_to_similarity(normed_feature: Tensor, label: Tensor) -> Tuple[Tensor, Tensor]:
#      similarity_matrix = normed_feature @ normed_feature.transpose(1, 0)
#      label_matrix = label.unsqueeze(1) == label.unsqueeze(0)
#
#      positive_matrix = label_matrix.triu(diagonal=1)
#      negative_matrix = label_matrix.logical_not().triu(diagonal=1)
#
#      similarity_matrix = similarity_matrix.view(-1)
#      positive_matrix = positive_matrix.view(-1)
#      negative_matrix = negative_matrix.view(-1)
#      return similarity_matrix[positive_matrix], similarity_matrix[negative_matrix]
#
#
#  class CircleLoss(nn.Module):
#      def __init__(self, m: float, gamma: float) -> None:
#          super(CircleLoss, self).__init__()
#          self.m = m
#          self.gamma = gamma
#          self.soft_plus = nn.Softplus()
#
#      def forward(self, sp: Tensor, sn: Tensor) -> Tensor:
#          sp = nn.functional.normalize(sp)
#          sp, sn = convert_label_to_similarity(sp, sn)
#          ap = torch.clamp_min(- sp.detach() + 1 + self.m, min=0.)
#          an = torch.clamp_min(sn.detach() + self.m, min=0.)
#
#          delta_p = 1 - self.m
#          delta_n = self.m
#
#          logit_p = - ap * (sp - delta_p) * self.gamma
#          logit_n = an * (sn - delta_n) * self.gamma
#
#          loss = self.soft_plus(torch.logsumexp(logit_n, dim=0) + torch.logsumexp(logit_p, dim=0))
#
#          return loss



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import tensorflow as tf
from keras import backend as K
from keras.legacy import interfaces
from keras.optimizers import Optimizer
import keras.backend as K
import numpy as np
#  from .utils import _apply_weight_decays, _compute_eta_t
#  from .utils import _apply_lr_multiplier, _check_args, K_eval

import random
#  from termcolor import colored
'''Helper methods for optimizers
'''


def warn_str():
    #  return colored('WARNING: ', 'red')
    return ""


def get_weight_decays(model, verbose=1):
    wd_dict = {}
    for layer in model.layers:
        layer_l2regs = _get_layer_l2regs(layer)
        if layer_l2regs:
            for layer_l2 in layer_l2regs:
                weight_name, weight_l2 = layer_l2
                wd_dict.update({weight_name: weight_l2})
                if weight_l2 != 0 and verbose:
                    print(
                        (warn_str() + "{} l2-regularization = {} - should be "
                         "set 0 before compiling model").format(
                             weight_name, weight_l2))
    return wd_dict


def fill_dict_in_order(_dict, _list_of_vals):
    for idx, key in enumerate(_dict.keys()):
        _dict[key] = _list_of_vals[idx]
    return _dict


def _get_layer_l2regs(layer):
    if hasattr(layer, 'cell') or \
      (hasattr(layer, 'layer') and hasattr(layer.layer, 'cell')):
        return _rnn_l2regs(layer)
    elif hasattr(layer, 'layer') and not hasattr(layer.layer, 'cell'):
        layer = layer.layer
    l2_lambda_kb = []
    for weight_name in ['kernel', 'bias']:
        _lambda = getattr(layer, weight_name + '_regularizer', None)
        if _lambda is not None:
            l2_lambda_kb.append(
                [getattr(layer, weight_name).name,
                 float(_lambda.l2)])
    return l2_lambda_kb


def _rnn_l2regs(layer):
    l2_lambda_krb = []
    if hasattr(layer, 'backward_layer'):
        for layer in [layer.forward_layer, layer.backward_layer]:
            l2_lambda_krb += _cell_l2regs(layer.cell)
        return l2_lambda_krb
    else:
        return _cell_l2regs(layer.cell)


def _cell_l2regs(rnn_cell):
    cell = rnn_cell
    l2_lambda_krb = []  # kernel-recurrent-bias

    for weight_idx, weight_type in enumerate(['kernel', 'recurrent', 'bias']):
        _lambda = getattr(cell, weight_type + '_regularizer', None)
        if _lambda is not None:
            weight_name = cell.weights[weight_idx].name
            l2_lambda_krb.append([weight_name, float(_lambda.l2)])
    return l2_lambda_krb


def _apply_weight_decays(cls, var, var_t):
    wd = cls.weight_decays[var.name]
    wd_normalized = wd * K.cast(
        K.sqrt(cls.batch_size / cls.total_iterations_wd), 'float32')
    var_t = var_t - cls.eta_t * wd_normalized * var

    if cls.init_verbose and not cls._init_notified:
        print('{} weight decay set for {}'.format(K_eval(wd_normalized),
                                                  var.name))
    return var_t


def _compute_eta_t(cls):
    PI = 3.141592653589793
    t_frac = K.cast(cls.t_cur / cls.total_iterations, 'float32')
    eta_t = cls.eta_min + 0.5 * (cls.eta_max - cls.eta_min) * \
        (1 + K.cos(PI * t_frac))
    return eta_t


def _apply_lr_multiplier(cls, lr_t, var):
    multiplier_name = [
        mult_name for mult_name in cls.lr_multipliers if mult_name in var.name
    ]
    if multiplier_name != []:
        lr_mult = cls.lr_multipliers[multiplier_name[0]]
    else:
        lr_mult = 1
    lr_t = lr_t * lr_mult

    if cls.init_verbose and not cls._init_notified:
        if lr_mult != 1:
            print('{} init learning rate set for {} -- {}'.format(
                '%.e' % K_eval(lr_t), var.name, lr_t))
        else:
            print('No change in learning rate {} -- {}'.format(
                var.name, K_eval(lr_t)))
    return lr_t


def _check_args(total_iterations, use_cosine_annealing, weight_decays):
    if use_cosine_annealing and total_iterations != 0:
        print('Using cosine annealing learning rates')
    elif (use_cosine_annealing
          or weight_decays != {}) and total_iterations == 0:
        print(warn_str() + "'total_iterations'==0, must be !=0 to use " +
              "cosine annealing and/or weight decays; " +
              "proceeding without either")


def reset_seeds(reset_graph_with_backend=None, verbose=1):
    if reset_graph_with_backend is not None:
        K = reset_graph_with_backend
        K.clear_session()
        tf.compat.v1.reset_default_graph()
        if verbose:
            print("KERAS AND TENSORFLOW GRAPHS RESET")

    np.random.seed(1)
    random.seed(2)
    if tf.__version__[0] == '2':
        tf.random.set_seed(3)
    else:
        tf.set_random_seed(3)
    if verbose:
        print("RANDOM SEEDS RESET")


def K_eval(x, backend=K):
    K = backend
    try:
        return K.get_value(K.to_dense(x))
    except Exception as e:
        try:
            eval_fn = K.function([], [x])
            return eval_fn([])[0]
        except Exception as e:
            return K.eager(K.eval)(x)


class AdamW(Optimizer):
    """AdamW optimizer.
    Default parameters follow those provided in the original paper.
    # Arguments
        learning_rate: float >= 0. Learning rate.
        beta_1: float, 0 < beta < 1. Generally close to 1.
        beta_2: float, 0 < beta < 1. Generally close to 1.
        amsgrad: boolean. Whether to apply the AMSGrad variant of this
            algorithm from the paper "On the Convergence of Adam and Beyond".
        batch_size:       int >= 1. Train input batch size; used for normalization
        total_iterations: int >= 0. Total expected iterations / weight updates
                          throughout training, used for normalization; <1>
        weight_decays:    dict / None. Name-value pairs specifying weight decays,
                          as {<weight matrix name>:<weight decay value>}; <2>
        lr_multipliers:   dict / None. Name-value pairs specifying per-layer lr
                          multipliers, as {<layer name>:<multiplier value>}; <2>
        use_cosine_annealing: bool. If True, multiplies lr each train iteration
                              as a function of eta_min, eta_max, total_iterations,
                              and t_cur (current); [2]-Appendix, 2
        eta_min, eta_max: int, int. Min & max values of cosine annealing
                          lr multiplier; [2]-Appendix, 2
        t_cur: int. Value to initialize t_cur to - used for 'warm restarts'.
               To be used together with use_cosine_annealing==True
        total_iterations_wd: int / None. If not None, weight_decays will be
                     applied according to total_iterations_wd instead of
                     total_iterations, contrary to authors' scheme. Set to
                     sum(total_iterations) over all restarts to normalize over
                     all epochs. May yield improvement over `None`.
        init_verbose: bool. If True, print weight-name--weight-decay, and
                      lr-multiplier--layer-name value pairs set during
                      optimizer initialization (recommended)
    # <1> - if using 'warm restarts', then refers to total expected iterations
            for a given restart; can be an estimate, and training won't stop
            at iterations == total_iterations. [2]-Appendix, pg 1
    # <2> - [AdamW Keras Implementation - Github repository]
            (https://github.com/OverLordGoldDragon/keras_adamw)
    # References
        - [1][Adam - A Method for Stochastic Optimization]
             (http://arxiv.org/abs/1412.6980v8)
        - [2][Fixing Weight Decay Regularization in Adam]
             (https://arxiv.org/abs/1711.05101)
    """
    def __init__(self,
                 learning_rate=0.001,
                 beta_1=0.9,
                 beta_2=0.999,
                 amsgrad=False,
                 batch_size=32,
                 total_iterations=0,
                 total_iterations_wd=None,
                 use_cosine_annealing=False,
                 weight_decays=None,
                 lr_multipliers=None,
                 init_verbose=True,
                 eta_min=0,
                 eta_max=1,
                 t_cur=0,
                 **kwargs):
        self.initial_decay = kwargs.pop('decay', 0.0)
        self.epsilon = kwargs.pop('epsilon', K.epsilon())
        learning_rate = kwargs.pop('lr', learning_rate)
        eta_t = kwargs.pop('eta_t', 1.)
        super(AdamW, self).__init__(**kwargs)

        with K.name_scope(self.__class__.__name__):
            self.iterations = K.variable(0, dtype='int64', name='iterations')
            self.learning_rate = K.variable(learning_rate,
                                            name='learning_rate')
            self.beta_1 = K.variable(beta_1, name='beta_1')
            self.beta_2 = K.variable(beta_2, name='beta_2')
            self.decay = K.variable(self.initial_decay, name='decay')
            self.batch_size = K.variable(batch_size,
                                         dtype='int64',
                                         name='batch_size')
            self.eta_min = K.constant(eta_min, name='eta_min')
            self.eta_max = K.constant(eta_max, name='eta_max')
            self.eta_t = K.variable(eta_t, dtype='float32', name='eta_t')
            self.t_cur = K.variable(t_cur, dtype='int64', name='t_cur')

        self.total_iterations = total_iterations
        self.total_iterations_wd = total_iterations_wd or total_iterations
        self.amsgrad = amsgrad
        self.lr_multipliers = lr_multipliers
        self.weight_decays = weight_decays or {}
        self.init_verbose = init_verbose
        self.use_cosine_annealing = use_cosine_annealing

        self._init_notified = False
        _check_args(total_iterations, use_cosine_annealing, self.weight_decays)

    @interfaces.legacy_get_updates_support
    @K.symbolic
    def get_updates(self, loss, params):
        grads = self.get_gradients(loss, params)
        self.updates = [K.update_add(self.iterations, 1)]
        self.updates.append(K.update_add(self.t_cur, 1))

        lr = self.learning_rate
        if self.initial_decay > 0:
            lr = lr * (1. / (1. + self.decay *
                             K.cast(self.iterations, K.dtype(self.decay))))

        t = K.cast(self.iterations, K.floatx()) + 1
        lr_t = lr * (K.sqrt(1. - K.pow(self.beta_2, t)) /
                     (1. - K.pow(self.beta_1, t)))

        ms = [
            K.zeros(K.int_shape(p), dtype=K.dtype(p), name='m_' + str(i))
            for (i, p) in enumerate(params)
        ]
        vs = [
            K.zeros(K.int_shape(p), dtype=K.dtype(p), name='v_' + str(i))
            for (i, p) in enumerate(params)
        ]

        if self.amsgrad:
            vhats = [
                K.zeros(K.int_shape(p),
                        dtype=K.dtype(p),
                        name='vhat_' + str(i)) for (i, p) in enumerate(params)
            ]
        else:
            vhats = [
                K.zeros(1, name='vhat_' + str(i)) for i in range(len(params))
            ]
        self.weights = [self.iterations] + ms + vs + vhats

        total_iterations = self.total_iterations
        # Cosine annealing
        if self.use_cosine_annealing and total_iterations != 0:
            self.eta_t = _compute_eta_t(self)
        self.lr_t = lr_t * self.eta_t  # for external tracking

        for p, g, m, v, vhat in zip(params, grads, ms, vs, vhats):
            # Learning rate multipliers
            if self.lr_multipliers is not None:
                lr_t = _apply_lr_multiplier(self, lr_t, p)

            m_t = (self.beta_1 * m) + (1. - self.beta_1) * g
            v_t = (self.beta_2 * v) + (1. - self.beta_2) * K.square(g)
            if self.amsgrad:
                vhat_t = K.maximum(vhat, v_t)
                p_t = p - lr_t * m_t / (K.sqrt(vhat_t) + self.epsilon)
                self.updates.append(K.update(vhat, vhat_t))
            else:
                p_t = p - lr_t * m_t / (K.sqrt(v_t) + self.epsilon)

            self.updates.append(K.update(m, m_t))
            self.updates.append(K.update(v, v_t))

            # Weight decays
            if p.name in self.weight_decays.keys() and total_iterations != 0:
                p_t = _apply_weight_decays(self, p, p_t)
            new_p = p_t

            # Apply constraints.
            if getattr(p, 'constraint', None) is not None:
                new_p = p.constraint(new_p)

            self.updates.append(K.update(p, new_p))

        self._init_notified = True
        return self.updates

    def get_config(self):
        config = {
            'learning_rate': float(K_eval(self.learning_rate)),
            'beta_1': float(K_eval(self.beta_1)),
            'beta_2': float(K_eval(self.beta_2)),
            'decay': float(K_eval(self.decay)),
            'batch_size': int(K_eval(self.batch_size)),
            'total_iterations': int(self.total_iterations),
            'weight_decays': self.weight_decays,
            'lr_multipliers': self.lr_multipliers,
            'use_cosine_annealing': self.use_cosine_annealing,
            't_cur': int(K_eval(self.t_cur)),
            'eta_t': float(K_eval(self.eta_t)),
            'eta_min': float(K_eval(self.eta_min)),
            'eta_max': float(K_eval(self.eta_max)),
            'init_verbose': self.init_verbose,
            'epsilon': self.epsilon,
            'amsgrad': self.amsgrad
        }
        base_config = super(AdamW, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))


class NadamW(Optimizer):
    """Nesterov Adam optimizer.
    Much like Adam is essentially RMSprop with momentum,
    Nadam is Adam RMSprop with Nesterov momentum.
    Default parameters follow those provided in the paper.
    It is recommended to leave the parameters of this optimizer
    at their default values.
    # Arguments
        lr: float >= 0. Learning rate.
        beta_1/beta_2: floats, 0 < beta < 1. Generally close to 1.
        epsilon: float >= 0. Fuzz factor. If `None`, defaults to `K.epsilon()`.
    # Arguments (other): see AdamW
    # References
        - [Nadam report](http://cs229.stanford.edu/proj2015/054_report.pdf)
        - [On the importance of initialization and momentum in deep learning]
          (http://www.cs.toronto.edu/~fritz/absps/momentum.pdf)
    """
    def __init__(self,
                 learning_rate=0.002,
                 beta_1=0.9,
                 beta_2=0.999,
                 batch_size=32,
                 total_iterations=0,
                 total_iterations_wd=None,
                 use_cosine_annealing=False,
                 weight_decays=None,
                 lr_multipliers=None,
                 init_verbose=True,
                 eta_min=0,
                 eta_max=1,
                 t_cur=0,
                 **kwargs):
        self.schedule_decay = kwargs.pop('schedule_decay', 0.004)
        self.epsilon = kwargs.pop('epsilon', K.epsilon())
        learning_rate = kwargs.pop('lr', learning_rate)
        eta_t = kwargs.pop('eta_t', 1.)
        super(NadamW, self).__init__(**kwargs)

        with K.name_scope(self.__class__.__name__):
            self.iterations = K.variable(0, dtype='int64', name='iterations')
            self.m_schedule = K.variable(1., name='m_schedule')
            self.learning_rate = K.variable(learning_rate,
                                            name='learning_rate')
            self.beta_1 = K.variable(beta_1, name='beta_1')
            self.beta_2 = K.variable(beta_2, name='beta_2')
            self.batch_size = K.variable(batch_size,
                                         dtype='int64',
                                         name='batch_size')
            self.eta_min = K.constant(eta_min, name='eta_min')
            self.eta_max = K.constant(eta_max, name='eta_max')
            self.eta_t = K.variable(eta_t, dtype='float32', name='eta_t')
            self.t_cur = K.variable(t_cur, dtype='int64', name='t_cur')

        self.total_iterations = total_iterations
        self.total_iterations_wd = total_iterations_wd or total_iterations
        self.lr_multipliers = lr_multipliers
        self.weight_decays = weight_decays or {}
        self.use_cosine_annealing = use_cosine_annealing
        self.init_verbose = init_verbose

        self._init_notified = False
        _check_args(total_iterations, use_cosine_annealing, self.weight_decays)

    @interfaces.legacy_get_updates_support
    @K.symbolic
    def get_updates(self, loss, params):
        grads = self.get_gradients(loss, params)
        self.updates = [K.update_add(self.iterations, 1)]
        self.updates.append(K.update_add(self.t_cur, 1))

        t = K.cast(self.iterations, K.floatx()) + 1

        # Due to the recommendations in [2], i.e. warming momentum schedule
        momentum_cache_t = self.beta_1 * (
            1. - 0.5 *
            (K.pow(K.cast_to_floatx(0.96), t * self.schedule_decay)))
        momentum_cache_t_1 = self.beta_1 * (
            1. - 0.5 * (K.pow(K.cast_to_floatx(0.96),
                              (t + 1) * self.schedule_decay)))
        m_schedule_new = self.m_schedule * momentum_cache_t
        m_schedule_next = self.m_schedule * momentum_cache_t * momentum_cache_t_1
        self.updates.append((self.m_schedule, m_schedule_new))

        shapes = [K.int_shape(p) for p in params]
        ms = [
            K.zeros(shape, name='m_' + str(i))
            for (i, shape) in enumerate(shapes)
        ]
        vs = [
            K.zeros(shape, name='v_' + str(i))
            for (i, shape) in enumerate(shapes)
        ]

        self.weights = [self.iterations, self.m_schedule] + ms + vs

        total_iterations = self.total_iterations
        # Cosine annealing
        if self.use_cosine_annealing and total_iterations != 0:
            self.eta_t = _compute_eta_t(self)
        self.lr_t = self.learning_rate * self.eta_t  # for external tracking

        for p, g, m, v in zip(params, grads, ms, vs):
            # Learning rate multipliers
            lr_t = self.learning_rate
            if self.lr_multipliers is not None:
                lr_t = _apply_lr_multiplier(self, lr_t, p)

            # the following equations given in [1]
            g_prime = g / (1. - m_schedule_new)
            m_t = self.beta_1 * m + (1. - self.beta_1) * g
            m_t_prime = m_t / (1. - m_schedule_next)
            v_t = self.beta_2 * v + (1. - self.beta_2) * K.square(g)
            v_t_prime = v_t / (1. - K.pow(self.beta_2, t))
            m_t_bar = (1. - momentum_cache_t) * g_prime + (momentum_cache_t_1 *
                                                           m_t_prime)

            self.updates.append(K.update(m, m_t))
            self.updates.append(K.update(v, v_t))
            p_t = p - self.eta_t * lr_t * m_t_bar / (K.sqrt(v_t_prime) +
                                                     self.epsilon)

            # Weight decays
            if p.name in self.weight_decays.keys() and total_iterations != 0:
                p_t = _apply_weight_decays(self, p, p_t)
            new_p = p_t

            # Apply constraints.
            if getattr(p, 'constraint', None) is not None:
                new_p = p.constraint(new_p)

            self.updates.append(K.update(p, new_p))

        self._init_notified = True
        return self.updates

    def set_weights(self, weights):
        params = self.weights
        # Override set_weights for backward compatibility of Keras 2.2.4 optimizer
        # since it does not include m_schedule at head of the weight list. Set
        # m_schedule to 1.
        if len(params) == len(weights) + 1:
            weights = [weights[0]] + [np.array(1.)] + weights[1:]
        super(NadamW, self).set_weights(weights)

    def get_config(self):
        config = {
            'learning_rate': float(K_eval(self.learning_rate)),
            'beta_1': float(K_eval(self.beta_1)),
            'beta_2': float(K_eval(self.beta_2)),
            'epsilon': self.epsilon,
            'schedule_decay': self.schedule_decay,
            'batch_size': int(K_eval(self.batch_size)),
            'total_iterations': int(self.total_iterations),
            'weight_decays': self.weight_decays,
            'lr_multipliers': self.lr_multipliers,
            'use_cosine_annealing': self.use_cosine_annealing,
            't_cur': int(K_eval(self.t_cur)),
            'eta_t': float(K_eval(self.eta_t)),
            'eta_min': float(K_eval(self.eta_min)),
            'eta_max': float(K_eval(self.eta_max)),
            'init_verbose': self.init_verbose
        }
        base_config = super(NadamW, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))


class SGDW(Optimizer):
    """Stochastic gradient descent optimizer.
    Includes support for momentum,
    learning rate decay, and Nesterov momentum.
    # Arguments
        lr: float >= 0. Learning rate.
        momentum: float >= 0. Parameter that accelerates SGD
            in the relevant direction and dampens oscillations.
        decay: float >= 0. Learning rate decay over each update.
        nesterov: boolean. Whether to apply Nesterov momentum.
    # Arguments (other): see AdamW
    """
    def __init__(self,
                 learning_rate=0.01,
                 momentum=0.,
                 nesterov=False,
                 batch_size=32,
                 total_iterations=0,
                 total_iterations_wd=None,
                 use_cosine_annealing=False,
                 weight_decays=None,
                 lr_multipliers=None,
                 init_verbose=True,
                 eta_min=0,
                 eta_max=1,
                 t_cur=0,
                 **kwargs):
        self.initial_decay = kwargs.pop('decay', 0.0)
        learning_rate = kwargs.pop('lr', learning_rate)
        eta_t = kwargs.pop('eta_t', 1.)
        super(SGDW, self).__init__(**kwargs)

        with K.name_scope(self.__class__.__name__):
            self.iterations = K.variable(0, dtype='int64', name='iterations')
            self.learning_rate = K.variable(learning_rate,
                                            name='learning_rate')
            self.momentum = K.variable(momentum, name='momentum')
            self.decay = K.variable(self.initial_decay, name='decay')
            self.batch_size = K.variable(batch_size,
                                         dtype='int64',
                                         name='batch_size')
            self.eta_min = K.constant(eta_min, name='eta_min')
            self.eta_max = K.constant(eta_max, name='eta_max')
            self.eta_t = K.variable(eta_t, dtype='float32', name='eta_t')
            self.t_cur = K.variable(t_cur, dtype='int64', name='t_cur')

        self.total_iterations = total_iterations
        self.total_iterations_wd = total_iterations_wd or total_iterations
        self.nesterov = nesterov
        self.lr_multipliers = lr_multipliers
        self.weight_decays = weight_decays or {}
        self.init_verbose = init_verbose
        self.use_cosine_annealing = use_cosine_annealing

        self._init_notified = False
        _check_args(total_iterations, use_cosine_annealing, self.weight_decays)

    @interfaces.legacy_get_updates_support
    @K.symbolic
    def get_updates(self, loss, params):
        grads = self.get_gradients(loss, params)
        self.updates = [K.update_add(self.iterations, 1)]
        self.updates.append(K.update_add(self.t_cur, 1))

        lr = self.learning_rate
        if self.initial_decay > 0:
            lr = lr * (1. / (1. + self.decay *
                             K.cast(self.iterations, K.dtype(self.decay))))
        # momentum
        shapes = [K.int_shape(p) for p in params]
        moments = [
            K.zeros(shape, name='moment_' + str(i))
            for (i, shape) in enumerate(shapes)
        ]
        self.weights = [self.iterations] + moments

        total_iterations = self.total_iterations
        # Cosine annealing
        if self.use_cosine_annealing and total_iterations != 0:
            self.eta_t = _compute_eta_t(self)
        self.lr_t = lr * self.eta_t  # for external tracking

        for p, g, m in zip(params, grads, moments):
            # Learning rate multipliers
            lr_t = self.learning_rate
            if self.lr_multipliers is not None:
                lr_t = _apply_lr_multiplier(self, lr_t, p)

            v = self.momentum * m - self.eta_t * lr_t * g  # velocity
            self.updates.append(K.update(m, v))

            if self.nesterov:
                p_t = p + self.momentum * v - self.eta_t * lr_t * g
            else:
                p_t = p + v

            # Weight decays
            if p.name in self.weight_decays.keys() and total_iterations != 0:
                p_t = _apply_weight_decays(self, p, p_t)
            new_p = p_t

            # Apply constraints.
            if getattr(p, 'constraint', None) is not None:
                new_p = p.constraint(new_p)

            self.updates.append(K.update(p, new_p))

        self._init_notified = True
        return self.updates

    def get_config(self):
        config = {
            'learning_rate': float(K_eval(self.learning_rate)),
            'momentum': float(K_eval(self.momentum)),
            'decay': float(K_eval(self.decay)),
            'nesterov': self.nesterov,
            'batch_size': int(K_eval(self.batch_size)),
            'total_iterations': int(self.total_iterations),
            'weight_decays': self.weight_decays,
            'lr_multipliers': self.lr_multipliers,
            'use_cosine_annealing': self.use_cosine_annealing,
            't_cur': int(K_eval(self.t_cur)),
            'eta_t': float(K_eval(self.eta_t)),
            'eta_min': float(K_eval(self.eta_min)),
            'eta_max': float(K_eval(self.eta_max)),
            'init_verbose': self.init_verbose
        }
        base_config = super(SGDW, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))



import torch
from torch import nn

class DiceLoss(nn.Module):
    """DiceLoss implemented from 'Dice Loss for Data-imbalanced NLP Tasks'
    Useful in dealing with unbalanced data
    """
    def __init__(self):
        super(DiceLoss, self).__init__()

    def forward(self,input, target):
        '''
        input: [N, C]
        target: [N, ]
        '''
        prob = torch.softmax(input, dim=1)
        prob = torch.gather(prob, dim=1, index=target.unsqueeze(1))
        dsc_i = 1 - ((1 - prob) * prob) / ((1 - prob) * prob + 1)
        dice_loss = dsc_i.mean()
        return dice_loss






# -*- coding: utf-8 -*-

import os, glob
import numpy as np
import pandas as pd

def remove_tail(s, t):
    i0 = s.find(t)
    if i0 >= 0:
        s = s[:i0]
    return s


def remove_head(s, h):
    i0 = s.find(h)
    if i0 >= 0:
        s = s[i0 + len(h):]
    return s



def get_same_rows(pd_first: pd.DataFrame, pd_second: pd.DataFrame) -> list:
    same_rows = []
    for i, row in pd_first.iterrows():
        if row.PREDICTION == pd_second.iloc[i, :].PREDICTION:
            same_rows.append(row.PREDICTION)
        else:
            same_rows.append(np.nan)
    return pd.DataFrame(same_rows, columns=['subject'])


def remove_nan_rows(data: pd.DataFrame,
                    colname: str,
                    nan_token: str = "__nan__") -> (pd.DataFrame, int):
    num_rows = data.shape[0]
    data[colname] = data[colname].fillna(nan_token)
    nan_list = data[data[colname] == nan_token].index.tolist()
    data = data.drop(nan_list)
    num_removed = num_rows - data.shape[0]
    return data, num_removed


def mix_train_data(train_data, test_data, predicted_results, best_results):
    guess_data = test_data
    guess_data.subject = get_same_rows(predicted_results, best_results)
    mixed_train_data = pd.concat([train_data, guess_data])
    mixed_train_data, num_removed = remove_nan_rows(mixed_train_data,
                                                    "subject")
    print(f"{num_removed} nan rows in mixed_train_data have been removed.")
    return mixed_train_data


def ensure_empty_directory(target_dir):
    import os, glob
    if not os.path.isdir(target_dir):
        os.makedirs(target_dir)
    if len(glob.glob(f"target_dir/*")) > 0:
        logger.warning(f"{target_dir} must be empty")
        return False
    return True


def list_find(list1, list2):
    """在list1中寻找子串list2，如果找到，返回第一个下标；
    如果找不到，返回-1。
    """
    n_list2 = len(list2)
    for i in range(len(list1)):
        if list1[i:i + n_list2] == list2:
            return i
    return -1


def my_longest_common_substring(s1: str, s2: str) -> float:
    if s1 == s2:
        return 1.0
    c1 = {}
    c2 = {}
    for c in s1:
        c1[c] = c1.get(c, 0) + 1
    for c in s2:
        c2[c] = c2.get(c, 0) + 1
    s1_total_chars = len(s1)
    s2_total_chars = len(s2)

    TP = 0
    TN = 0
    FP = 0
    FN = 0

    for c, n1 in c1.items():
        n2 = c2.get(c, 0)
        TP += min(n1, n2)
        TN += n2 - min(n1, n2)
        FP += n1 - min(n1, n2)
    for c, n2 in c2.items():
        if c not in c1:
            TN += n2

    acc = (TP + FN) / s2_total_chars
    recall = (TP + FN) / s1_total_chars
    f1 = 2.0 * acc * recall / (acc + recall)

    return acc, recall, f1


def longest_common_substring(s1, s2):
    m = [[0] * (1 + len(s2)) for i in range(1 + len(s1))]
    longest, s1_end, s2_end = 0, 0, 0
    for x in range(1, 1 + len(s1)):
        for y in range(1, 1 + len(s2)):
            if s1[x - 1] == s2[y - 1]:
                m[x][y] = m[x - 1][y - 1] + 1
                if m[x][y] > longest:
                    longest = m[x][y]
                    s1_end = x
                    s2_end = y
            else:
                m[x][y] = 0
    return s1_end, s2_end, longest, s1[s1_end - longest:s1_end]


def merge_sentence(s1, s2):
    s1_words = s1.split(' ')
    s2_words = s2.split(' ')

    s1_end, s2_end, longest, substring = longest_common_substring(
        s1_words, s2_words)

    if len(substring) != 0:
        head = s1_words[:s1_end - longest]
        tail = s2_words[s2_end:]
        new_sent = head + substring + tail
        return ' '.join(new_sent)
    else:
        return ''


#  if __name__ == '__main__':
def test_longest_common_substring():
    #  s1 = 'we are interested in evaluating two closely related variants. One is a long short-term memory (LSTM) unit, and the other is a gated recurrent unit (GRU) proposed more recently by Cho et al. [2014].'
    #  s2 = 'proposed more recently by Cho et al. [2014]. It is well established in the field that the LSTM unit works well on sequence-based tasks with long-term dependencies, but the latter has only recently been introduced and used in the context of machine translation.'
    #
    #  s1_words = s1.split(' ')
    #  s2_words = s2.split(' ')
    s1 = "中国要伟大的"
    s2 = "美国也要伟大才行"
    s1_words = [w for w in s1]
    s2_words = [w for w in s2]

    # pass the words lists into lcs function,
    # so it would compare 2 sentences word by word instead of character by character.
    s1_end, s2_end, longest, substring = longest_common_substring(
        s1_words, s2_words)
    print(f'find common substring in s1[{s1_end - longest}:{s1_end}]')
    print(f'find common substring in s2[{s2_end - longest}:{s2_end}]')
    print(f'common substring = {substring}')

    # if there is a common part in these 2 sentences,
    # maybe you want to merge them into 1 sentence.
    merged = merge_sentence(s1, s2)
    if len(merged) != 0:
        print(f'merged 2 strings into: {merged}')


def test():
    s1 = "中国要伟大的"
    s2 = "美国也要伟大才行"
    acc, recall, f1 = my_longest_common_substring(s1, s2)
    print(f"s1: {s1}")
    print(f"s2: {s2}")
    print(f"f1: {f1:.3f}, acc: {acc:.3f}, recall: {recall:.3f}")


if __name__ == '__main__':
    #  test()
    test_longest_common_substring()



# -*- coding: utf-8 -*-



def enable_gpu_growth():
    pass
    #  import tensorflow as tf
    #  import keras.backend as K
    #  config = tf.ConfigProto()
    #  config.gpu_options.allow_growth = True
    #  sess = tf.Session(config=config)
    #  K.set_session(sess)
    #

def get_gpu_info(gpu_id: int = 0):
    import pynvml
    pynvml.nvmlInit()
    handle = pynvml.nvmlDeviceGetHandleByIndex(gpu_id)
    meminfo = pynvml.nvmlDeviceGetMemoryInfo(handle)
    print(
        f"GPU {gpu_id} mem: total: {meminfo.total}, used: {meminfo.used}, free: {meminfo.free}"
    )
    return meminfo



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from .utils import init_theta, init_random, init_cuda, seg_generator
from .utils import softmax, sigmoid, load_json_file, show_statistical_distribution, dataframe_to_examples, acc_and_f1, simple_accuracy, to_numpy, create_logger
from .utils import batch_generator, slide_generator, split_train_eval_examples, shuffle_list, get_list_size, list_to_list
from .utils import get_pred_results_file, get_submission_file



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import copy, math
import numpy as np
from tqdm import tqdm
from loguru import logger
import torch
import torch.nn as nn
from transformers import BertConfig, BertTokenizer
from transformers import AutoConfig, AutoModelForSequenceClassification, AutoTokenizer, BertTokenizer, BertPreTrainedModel

from ..trainer import Trainer
from ...utils import softmax, acc_and_f1
from ...utils.multiprocesses import barrier_leader_process, barrier_member_processes


def gelu(x):
    """Implementation of the gelu activation function.
        For information: OpenAI GPT's gelu is slightly different (and gives
            slightly different results):
        0.5 * x * (1 + torch.tanh(math.sqrt(2 / math.pi) * (x + 0.044715 * torch.pow(x, 3))))
        Also see https://arxiv.org/abs/1606.08415
    """
    return x * 0.5 * (1.0 + torch.erf(x / math.sqrt(2.0)))


def swish(x):
    return x * torch.sigmoid(x)


ACT2FN = {"gelu": gelu, "relu": torch.nn.functional.relu, "swish": swish}


#  try:
#      from apex.normalization.fused_layer_norm import FusedLayerNorm as BertLayerNorm
#  except ImportError:
#      logger.info("Better speed can be achieved with apex installed from" +
#                  "https://www.github.com/nvidia/apex .")
#
class BertLayerNorm(nn.Module):
    def __init__(self, hidden_size, eps=1e-12):
        """Construct a layernorm module in the TF style
                (epsilon inside the square root).
        """
        super(BertLayerNorm, self).__init__()
        self.weight = nn.Parameter(torch.ones(hidden_size))
        self.bias = nn.Parameter(torch.zeros(hidden_size))
        self.variance_epsilon = eps

    def forward(self, x):
        u = x.mean(-1, keepdim=True)
        s = (x - u).pow(2).mean(-1, keepdim=True)
        x = (x - u) / torch.sqrt(s + self.variance_epsilon)
        return self.weight * x + self.bias


class BertSelfAttention(nn.Module):
    def __init__(self, config):
        super(BertSelfAttention, self).__init__()
        if config.hidden_size % config.num_attention_heads != 0:
            raise ValueError(
                "The hidden size (%d) is not a multiple of the number of attention "
                "heads (%d)" %
                (config.hidden_size, config.num_attention_heads))
        self.num_attention_heads = config.num_attention_heads
        self.attention_head_size = int(config.hidden_size /
                                       config.num_attention_heads)
        self.all_head_size = self.num_attention_heads * self.attention_head_size

        self.query = nn.Linear(config.hidden_size, self.all_head_size)
        self.key = nn.Linear(config.hidden_size, self.all_head_size)
        self.value = nn.Linear(config.hidden_size, self.all_head_size)

        self.dropout = nn.Dropout(config.attention_probs_dropout_prob)

    def transpose_for_scores(self, x):
        new_x_shape = x.size()[:-1] + (self.num_attention_heads,
                                       self.attention_head_size)
        x = x.view(*new_x_shape)
        return x.permute(0, 2, 1, 3)

    def forward(self, hidden_states, attention_mask):
        mixed_query_layer = self.query(hidden_states)
        mixed_key_layer = self.key(hidden_states)
        mixed_value_layer = self.value(hidden_states)

        query_layer = self.transpose_for_scores(mixed_query_layer)
        key_layer = self.transpose_for_scores(mixed_key_layer)
        value_layer = self.transpose_for_scores(mixed_value_layer)

        # Take the dot product between "query" and "key" to get the raw attention scores.
        attention_scores = torch.matmul(query_layer,
                                        key_layer.transpose(-1, -2))
        attention_scores = attention_scores / math.sqrt(
            self.attention_head_size)
        # Apply the attention mask is (precomputed for all layers in BertModel forward() function)
        attention_scores = attention_scores + attention_mask

        # Normalize the attention scores to probabilities.
        attention_probs = nn.Softmax(dim=-1)(attention_scores)

        # This is actually dropping out entire tokens to attend to, which might
        # seem a bit unusual, but is taken from the original Transformer paper.
        attention_probs = self.dropout(attention_probs)

        context_layer = torch.matmul(attention_probs, value_layer)
        context_layer = context_layer.permute(0, 2, 1, 3).contiguous()
        new_context_layer_shape = context_layer.size()[:-2] + (
            self.all_head_size, )
        context_layer = context_layer.view(*new_context_layer_shape)
        return context_layer


class BertSelfOutput(nn.Module):
    def __init__(self, config):
        super(BertSelfOutput, self).__init__()
        self.dense = nn.Linear(config.hidden_size, config.hidden_size)
        self.LayerNorm = BertLayerNorm(config.hidden_size, eps=1e-12)
        self.dropout = nn.Dropout(config.hidden_dropout_prob)

    def forward(self, hidden_states, input_tensor):
        hidden_states = self.dense(hidden_states)
        hidden_states = self.dropout(hidden_states)
        hidden_states = self.LayerNorm(hidden_states + input_tensor)
        return hidden_states


class BertAttention(nn.Module):
    def __init__(self, config):
        super(BertAttention, self).__init__()
        self.self = BertSelfAttention(config)
        self.output = BertSelfOutput(config)

    def forward(self, input_tensor, attention_mask):
        self_output = self.self(input_tensor, attention_mask)
        attention_output = self.output(self_output, input_tensor)
        return attention_output


class BertIntermediate(nn.Module):
    def __init__(self, config):
        super(BertIntermediate, self).__init__()
        self.dense = nn.Linear(config.hidden_size, config.intermediate_size)
        if isinstance(config.hidden_act, str):
            # or (sys.version_info[0] == 2 and isinstance(hidden_act, unicode)):
            self.intermediate_act_fn = ACT2FN[config.hidden_act]
        else:
            self.intermediate_act_fn = config.hidden_act

    def forward(self, hidden_states):
        hidden_states = self.dense(hidden_states)
        hidden_states = self.intermediate_act_fn(hidden_states)
        return hidden_states


class BertOutput(nn.Module):
    def __init__(self, config):
        super(BertOutput, self).__init__()
        self.dense = nn.Linear(config.intermediate_size, config.hidden_size)
        self.LayerNorm = BertLayerNorm(config.hidden_size, eps=1e-12)
        self.dropout = nn.Dropout(config.hidden_dropout_prob)

    def forward(self, hidden_states, input_tensor):
        hidden_states = self.dense(hidden_states)
        hidden_states = self.dropout(hidden_states)
        hidden_states = self.LayerNorm(hidden_states + input_tensor)
        return hidden_states


class BertLayer(nn.Module):
    def __init__(self, config):
        super(BertLayer, self).__init__()
        self.attention = BertAttention(config)
        self.intermediate = BertIntermediate(config)
        self.output = BertOutput(config)

    def forward(self, hidden_states, attention_mask):
        attention_output = self.attention(hidden_states, attention_mask)
        intermediate_output = self.intermediate(attention_output)
        layer_output = self.output(intermediate_output, attention_output)
        return layer_output


class BertEmbeddings(nn.Module):
    """Construct the embeddings from word, position and token_type embeddings.
    """
    def __init__(self, config):
        super(BertEmbeddings, self).__init__()
        self.word_embeddings = nn.Embedding(config.vocab_size,
                                            config.hidden_size,
                                            padding_idx=0)
        self.position_embeddings = nn.Embedding(config.max_position_embeddings,
                                                config.hidden_size)
        self.token_type_embeddings = nn.Embedding(config.type_vocab_size,
                                                  config.hidden_size)

        # self.LayerNorm is not snake-cased to stick with TensorFlow model
        #variable name and be able to load any TensorFlow checkpoint file
        self.LayerNorm = BertLayerNorm(config.hidden_size, eps=1e-12)
        self.dropout = nn.Dropout(config.hidden_dropout_prob)

    def forward(self, input_ids, token_type_ids=None):
        seq_length = input_ids.size(1)
        position_ids = torch.arange(seq_length,
                                    dtype=torch.long,
                                    device=input_ids.device)
        position_ids = position_ids.unsqueeze(0).expand_as(input_ids)
        if token_type_ids is None:
            token_type_ids = torch.zeros_like(input_ids)

        words_embeddings = self.word_embeddings(input_ids)
        position_embeddings = self.position_embeddings(position_ids)
        token_type_embeddings = self.token_type_embeddings(token_type_ids)

        embeddings = words_embeddings + position_embeddings + token_type_embeddings
        embeddings = self.LayerNorm(embeddings)
        embeddings = self.dropout(embeddings)
        return embeddings


class BertEncoder(nn.Module):
    def __init__(self, config):
        super(BertEncoder, self).__init__()
        layer = BertLayer(config)
        self.layer = nn.ModuleList(
            [copy.deepcopy(layer) for _ in range(config.num_hidden_layers)])

    def forward(self,
                hidden_states,
                attention_mask,
                output_all_encoded_layers=True):
        all_encoder_layers = []
        for layer_module in self.layer:
            hidden_states = layer_module(hidden_states, attention_mask)
            if output_all_encoded_layers:
                all_encoder_layers.append(hidden_states)
        if not output_all_encoded_layers:
            all_encoder_layers.append(hidden_states)
        return all_encoder_layers


class BertPooler(nn.Module):
    def __init__(self, config):
        super(BertPooler, self).__init__()
        self.dense = nn.Linear(config.hidden_size, config.hidden_size)
        self.activation = nn.Tanh()

    def forward(self, hidden_states):
        # We "pool" the model by simply taking the hidden state corresponding
        # to the first token.
        first_token_tensor = hidden_states[:, 0]
        pooled_output = self.dense(first_token_tensor)
        pooled_output = self.activation(pooled_output)
        return pooled_output


#  class BertPreTrainedModel(nn.Module):
#      """ An abstract class to handle weights initialization and
#          a simple interface for dowloading and loading pretrained models.
#      """
#      def __init__(self, config, *inputs, **kwargs):
#          super(BertPreTrainedModel, self).__init__()
#          if not isinstance(config, BertConfig):
#              raise ValueError(
#                  "Parameter config in `{}(config)` should be an instance of "
#                  "class `BertConfig`. "
#                  "To create a model from a Google pretrained model use "
#                  "`model = {}.from_pretrained(PRETRAINED_MODEL_NAME)`".format(
#                      self.__class__.__name__, self.__class__.__name__))
#          self.config = config
#
#      def init_bert_weights(self, module):
#          """ Initialize the weights.
#          """
#          if isinstance(module, (nn.Linear, nn.Embedding)):
#              # Slightly different from the TF version which uses
#              # truncated_normal for initialization
#              # cf https://github.com/pytorch/pytorch/pull/5617
#              module.weight.data.normal_(mean=0.0,
#                                         std=self.config.initializer_range)
#          elif isinstance(module, BertLayerNorm):
#              module.bias.data.zero_()
#              module.weight.data.fill_(1.0)
#          if isinstance(module, nn.Linear) and module.bias is not None:
#              module.bias.data.zero_()
#          return


class BertModel(BertPreTrainedModel):
    """BERT model ("Bidirectional Embedding Representations from a Transformer").

    Params:
        config: a BertConfig class instance with the configuration to build a new model

    Inputs:
        `input_ids`: a torch.LongTensor of shape [batch_size, sequence_length]
            with the word token indices in the vocabulary(see the tokens preprocessing
            logic in the scripts `extract_features.py`, `run_classifier.py` and
            `run_squad.py`)
        `token_type_ids`: an optional torch.LongTensor of shape
            [batch_size, sequence_length] with the token types indices selected
            in [0, 1]. Type 0 corresponds to a `sentence A` and type 1 corresponds to
            a `sentence B` token (see BERT paper for more details).
        `attention_mask`: an optional torch.LongTensor of shape [batch_size, sequence_length] with indices
            selected in [0, 1]. It's a mask to be used if the input sequence length is smaller than the max
            input sequence length in the current batch. It's the mask that we typically use for attention when
            a batch has varying length sentences.
        `output_all_encoded_layers`: boolean which controls the content of the `encoded_layers` output as described below. Default: `True`.

    Outputs: Tuple of (encoded_layers, pooled_output)
        `encoded_layers`: controled by `output_all_encoded_layers` argument:
            - `output_all_encoded_layers=True`: outputs a list of the full sequences of encoded-hidden-states at the end
                of each attention block (i.e. 12 full sequences for BERT-base, 24 for BERT-large), each
                encoded-hidden-state is a torch.FloatTensor of size [batch_size, sequence_length, hidden_size],
            - `output_all_encoded_layers=False`: outputs only the full sequence of hidden-states corresponding
                to the last attention block of shape [batch_size, sequence_length, hidden_size],
        `pooled_output`: a torch.FloatTensor of size [batch_size, hidden_size] which is the output of a
            classifier pretrained on top of the hidden state associated to the first character of the
            input (`CLS`) to train on the Next-Sentence task (see BERT's paper).

    Example usage:
    ```python
    # Already been converted into WordPiece token ids
    input_ids = torch.LongTensor([[31, 51, 99], [15, 5, 0]])
    input_mask = torch.LongTensor([[1, 1, 1], [1, 1, 0]])
    token_type_ids = torch.LongTensor([[0, 0, 1], [0, 1, 0]])

    config = modeling.BertConfig(vocab_size_or_config_json_file=32000, hidden_size=768,
        num_hidden_layers=12, num_attention_heads=12, intermediate_size=3072)

    model = modeling.BertModel(config=config)
    all_encoder_layers, pooled_output = model(input_ids, token_type_ids, input_mask)
    ```
    """
    def __init__(self, config):
        super(BertModel, self).__init__(config)
        self.embeddings = BertEmbeddings(config)
        self.encoder = BertEncoder(config)
        self.pooler = BertPooler(config)
        self.apply(self.init_bert_weights)

    def forward(self,
                input_ids,
                token_type_ids=None,
                attention_mask=None,
                output_all_encoded_layers=True):
        if attention_mask is None:
            attention_mask = torch.ones_like(input_ids)
        if token_type_ids is None:
            token_type_ids = torch.zeros_like(input_ids)

        # We create a 3D attention mask from a 2D tensor mask.
        # Sizes are [batch_size, 1, 1, to_seq_length]
        # So we can broadcast to [batch_size, num_heads, from_seq_length, to_seq_length]
        # this attention mask is more simple than the triangular masking of causal attention
        # used in OpenAI GPT, we just need to prepare the broadcast dimension here.
        extended_attention_mask = attention_mask.unsqueeze(1).unsqueeze(2)

        # Since attention_mask is 1.0 for positions we want to attend and 0.0 for
        # masked positions, this operation will create a tensor which is 0.0 for
        # positions we want to attend and -10000.0 for masked positions.
        # Since we are adding it to the raw scores before the softmax, this is
        # effectively the same as removing these entirely.
        extended_attention_mask = extended_attention_mask.to(
            dtype=next(self.parameters()).dtype)  # fp16 compatibility
        extended_attention_mask = (1.0 - extended_attention_mask) * -10000.0

        embedding_output = self.embeddings(input_ids, token_type_ids)
        encoded_layers = self.encoder(
            embedding_output,
            extended_attention_mask,
            output_all_encoded_layers=output_all_encoded_layers)
        sequence_output = encoded_layers[-1]
        pooled_output = self.pooler(sequence_output)
        if not output_all_encoded_layers:
            encoded_layers = encoded_layers[-1]
        return encoded_layers, pooled_output

    def init_bert_weights(self, module):
        """ Initialize the weights.
        """
        if isinstance(module, (nn.Linear, nn.Embedding)):
            # Slightly different from the TF version which uses
            # truncated_normal for initialization
            # cf https://github.com/pytorch/pytorch/pull/5617
            module.weight.data.normal_(mean=0.0,
                                       std=self.config.initializer_range)
        elif isinstance(module, BertLayerNorm):
            module.bias.data.zero_()
            module.weight.data.fill_(1.0)
        if isinstance(module, nn.Linear) and module.bias is not None:
            module.bias.data.zero_()
        return


class BertSentenceEmbedding(BertPreTrainedModel):
    """BERT model for sentence embedding.
    This module is composed of the BERT model with a linear layer on top of
    the pooled output. The output of the model is the embedding for the
    input sentence, which can be used for cosine similarity compuation.

    Params:
        `config`: a BertConfig class instance with the configuration to
                build a new model.
        `emb_size`: size of the sentence embedding vector.

    Inputs:
        `input_ids`: a torch.LongTensor of shape [batch_size, sequence_length]
            with the word token indices in the vocabulary.
        `attention_mask`: a torch.LongTensor of shape
            [batch_size, sequence_length] with indices selected in [0, 1].
            It's a mask to be used if the input sequence length is smaller than
            the max input sequence length in the current batch. It's the mask
            that we typically use for attention when a batch has varying length
            sentences.

    Outputs:
        Outputs the classification logits of shape [batch_size, emb_size].

    Example usage:
    ```python
    # Already been converted into WordPiece token ids
    input_ids = torch.LongTensor([[31, 51, 99], [15, 5, 0]])
    input_mask = torch.LongTensor([[1, 1, 1], [1, 1, 0]])

    config = BertConfig(vocab_size_or_config_json_file=32000, hidden_size=768,
        num_hidden_layers=12, num_attention_heads=12, intermediate_size=3072)

    emb_size = 1024

    model = BertSentenceEmbedding(config, emb_size)
    emb = model(input_ids, input_mask)
    ```
    """
    def __init__(self, config, emb_size):
        super(BertSentenceEmbedding, self).__init__(config)
        self.emb_size = emb_size
        self.bert = BertModel(config)
        self.emb = nn.Linear(config.hidden_size, emb_size)
        self.activation = nn.Tanh()
        self.cos_fn = nn.CosineSimilarity(dim=1, eps=1e-6)
        self.apply(self.init_bert_weights)

    def calcSim(self, emb1, emb2):
        return self.cos_fn(emb1, emb2)

    def forward(self, input_ids, attention_mask):
        _, pooled_output = self.bert(input_ids,
                                     None,
                                     attention_mask,
                                     output_all_encoded_layers=False)
        emb = self.activation(self.emb(pooled_output))
        return emb

    def init_bert_weights(self, module):
        """ Initialize the weights.
        """
        if isinstance(module, (nn.Linear, nn.Embedding)):
            # Slightly different from the TF version which uses
            # truncated_normal for initialization
            # cf https://github.com/pytorch/pytorch/pull/5617
            module.weight.data.normal_(mean=0.0,
                                       std=self.config.initializer_range)
        elif isinstance(module, BertLayerNorm):
            module.bias.data.zero_()
            module.weight.data.fill_(1.0)
        if isinstance(module, nn.Linear) and module.bias is not None:
            module.bias.data.zero_()
        return


class ClassReport(object):
    def __init__(self, target_names=None, labels=None):
        self.target_names = target_names
        self.labels = labels

    def __call__(self, output, target):
        _, y_pred = torch.max(output.data, 1)
        y_pred = y_pred.cpu().numpy()
        y_true = target.cpu().numpy()
        #  print(y_pred, y_true, self.labels, self.target_names)
        classify_report = classification_report(y_true,
                                                y_pred,
                                                labels=self.labels,
                                                target_names=self.target_names)
        print('\n\nclassify_report:\n', classify_report)


MODEL_CLASSES = {
    'bert': (AutoConfig, BertSentenceEmbedding, BertTokenizer),
}


def load_pretrained_tokenizer(args):
    config_class, model_class, tokenizer_class = MODEL_CLASSES[args.model_type]
    tokenizer = tokenizer_class.from_pretrained(
        args.model_path,
        do_lower_case=args.do_lower_case,
        cache_dir=args.cache_dir if args.cache_dir else None,
    )

    return tokenizer


def load_pretrained_model(args):
    # make sure only the first process in distributed training
    # will download model & vocab
    barrier_member_processes(args)

    config_class, model_class, tokenizer_class = MODEL_CLASSES[args.model_type]
    config = config_class.from_pretrained(
        args.model_path,
        num_labels=args.num_labels,
        cache_dir=args.cache_dir if args.cache_dir else None,
    )
    logger.info(f"model_path: {args.model_path}")
    logger.info(f"config:{config}")
    model = model_class.from_pretrained(
        args.model_path,
        from_tf=bool(".ckpt" in args.model_path),
        config=config,
        cache_dir=args.cache_dir if args.cache_dir else None,
        emb_size=1024)
    #  label2id=label2id,
    #  device=args.device)

    # make sure only the first process in distributed training
    # will download model & vocab
    barrier_leader_process(args)

    return model


class SenembTrainer(Trainer):
    def __init__(self):
        super(SenembTrainer, self).__init__()
        self.logits = None
        self.preds = None
        self.probs = None

    #  def package_inputs_from_batch(self, args, batch, known_labels=True):
    #      inputs = {
    #          "input_ids": batch[0],
    #          "attention_mask": batch[1],
    #          "labels": batch[3]
    #      }
    #      if args.model_type != "distilbert":
    #          # XLM, DistilBERT, RoBERTa, and XLM-RoBERTa don't use segment_ids
    #          inputs["token_type_ids"] = (batch[2] if args.model_type in [
    #              "bert", "xlnet", "albert"
    #          ] else None)
    #      return inputs

    def on_train_step(self, args, model, step, batch):
        def getLossFunction(args):
            cos_fn = nn.CosineSimilarity(dim=1, eps=1e-6).to(args.device)
            l1_loss = nn.L1Loss(reduction='mean')

            def cosine_loss(v1, v2, label):
                sim_t = cos_fn(v1, v2)
                d = l1_loss(sim_t, label)
                return d

            return cosine_loss

        input_ids1, input_mask1, input_ids2, input_mask2, label_ids = batch
        logits1 = model(input_ids1, input_mask1)
        logits2 = model(input_ids2, input_mask2)
        loss_fn = getLossFunction(args)
        loss = loss_fn(logits1, logits2, label_ids)

        return (loss, logits1, logits2)

    def on_eval_start(self, args, eval_dataset):
        self.logits = None
        self.preds = None
        self.out_label_ids = None
        self.results = {}
        self.probs = None

    #  def on_eval_step(self, args, eval_dataset, step, model, inputs, outputs):
    def on_eval_step(self, args, model, step, batch):
        input_ids1, input_mask1, input_ids2, input_mask2, label_ids = batch
        #  eval_loss, logits = outputs[:2]
        eval_loss, logits1, logits2 = self.on_train_step(
            args, model, step, batch)

        cos_fn = nn.CosineSimilarity(dim=1, eps=1e-6)
        logits = cos_fn(logits1, logits2)

        if self.logits is None:
            self.logits = logits.detach().cpu().numpy()
            self.out_label_ids = label_ids.detach().cpu().numpy()
        else:
            self.logits = np.append(self.logits,
                                    logits.detach().cpu().numpy(),
                                    axis=0)
            self.out_label_ids = np.append(self.out_label_ids,
                                           label_ids.detach().cpu().numpy(),
                                           axis=0)

        #  self.preds = np.argmax(self.logits, axis=1)
        #  self.probs = softmax(self.logits)
        self.preds = np.array([1 if x > 0.5 else 0 for x in self.logits])
        self.probs = np.array([x if x > 0.5 else 1 - x for x in self.logits])

        result = acc_and_f1(self.preds, self.out_label_ids)
        self.results.update(result)

        return (eval_loss, ), self.results

    def on_eval_end(self, args, eval_dataset):
        #  self.preds = np.argmax(self.preds, axis=1)
        # for regessions
        #  self.preds = np.squeeze(self.preds)

        logger.info(f"  Num examples = {len(eval_dataset)}")
        logger.info(f"  Batch size = {args.eval_batch_size}")
        logger.info(f"******** Eval results ********")
        for key in self.results.keys():
            logger.info(f" dev: {key} = {self.results[key]:.4f}")

        return self.results

    def on_predict_start(self, args, test_dataset):
        self.logits = None

        self.preds = None
        self.probs = None

    def on_predict_step(self, args, test_dataset, step, model, inputs,
                        outputs):
        _, logits = outputs[:2]
        if self.logits is None:
            self.logits = logits.detach().cpu().numpy()
        else:
            self.logits = np.append(self.logits,
                                    logits.detach().cpu().numpy(),
                                    axis=0)

    def on_predict_end(self, args, test_dataset):
        self.preds = np.argmax(self.logits, axis=1)
        self.probs = softmax(self.logits)
        logger.debug(self.preds)
        logger.debug(self.probs)

    def encode(self, args, model, encode_dataset):

        encode_dataloader = self.generate_dataloader(args,
                                                     encode_dataset,
                                                     batch_size=1,
                                                     keep_order=True)

        embeddings = None
        for step, batch in enumerate(encode_dataloader):
            model.eval()
            batch = tuple(t.to(args.device) for t in batch)
            with torch.no_grad():
                input_ids, input_mask = batch
                logits = model(input_ids, input_mask)

                if embeddings is None:
                    embeddings = logits.detach().cpu().numpy()
                else:
                    embeddings = np.append(embeddings,
                                           logits.detach().cpu().numpy(),
                                           axis=0)
        return embeddings



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os, json, copy
from pathlib import Path
from loguru import logger

import torch
from torch.utils.data import DataLoader, RandomSampler, SequentialSampler
from ..utils import init_theta

try:
    from torch.utils.tensorboard import SummaryWriter
except ImportError:
    from tensorboardX import SummaryWriter

from ..utils.progbar import Progbar
from ..utils.multiprocesses import is_master_process, is_multi_processes, barrier_leader_process, barrier_member_processes


def get_default_optimizer_parameters(model, weight_decay):
    param_optimizers = list(model.named_parameters())
    no_decay = ['bias', 'LayerNorm.weights']
    optimizer_grouped_parameters = [{
        'params': [
            p for n, p in param_optimizers
            if not any(nd in n for nd in no_decay)
        ],
        'weight_decay':
        weight_decay
    }, {
        'params':
        [p for n, p in param_optimizers if any(nd in n for nd in no_decay)],
        'weight_decay':
        0.0
    }]

    return optimizer_grouped_parameters


def generate_dataloader(args,
                        dataset,
                        batch_size,
                        keep_order=True,
                        collate_fn=None):

    Sampler = SequentialSampler if keep_order else RandomSampler
    sampler = DistributedSampler(dataset) if is_multi_processes(
        args) else Sampler(dataset)
    dataloader = DataLoader(dataset,
                            sampler=sampler,
                            batch_size=batch_size,
                            collate_fn=collate_fn)
    return dataloader


class Trainer:
    def __init__(self, args):
        init_theta(args)

        self.args = args
        self.collate_fn = None

    #  def build_model(self, args):
    #      raise NotImplementedError

    def batch_to_inputs(self, args, batch, known_labels=True):
        raise NotImplementedError

    def examples_to_dataset(self, examples, max_seq_length):
        raise NotImplementedError

    #  def generate_dataloader(self, args, dataset, batch_size, keep_order=True):
    #
    #      return generate_dataloader(args, dataset, batch_size, keep_order)

    def on_train_step(self, args, model, step, batch):
        inputs = self.batch_to_inputs(args, batch)
        outputs = model(**inputs)
        return outputs

    def on_eval_start(self, args, eval_dataset):
        pass

    #  def on_eval_step(self, args, eval_dataset, step, model, inputs, outputs):
    #      results = {}
    #      return results

    def on_eval_step(self, args, model, step, batch, batch_features):
        inputs = self.batch_to_inputs(args, batch)
        outputs = model(**inputs)
        return outputs, {}

    def on_eval_end(self, args, eval_dataset):
        results = {}
        return results

    def on_predict_start(self, args, test_dataset):
        pass

    #  def on_predict_step(self, args, test_dataset, step, model, inputs,
    #                      outputs):
    def on_predict_step(self, args, model, step, batch):
        pass

    def on_predict_end(self, args, test_dataset):
        return []

    def save_model(self, args, model, tokenizer, optimizer, scheduler,
                   model_path):
        if not os.path.exists(model_path):
            os.makedirs(model_path)

        logger.info(f"Saving model checkpoint to {model_path}")

        model_to_save = (model.module if hasattr(model, "module") else model)
        model_to_save.save_pretrained(model_path)

        tokenizer.save_vocabulary(Path(model_path).as_posix() + '/')

        #  torch.save(args, os.path.join(model_path, "training_args.bin"))
        json.dump(
            {
                k: v
                for k, v in args.__dict__.items() if v is None
                or type(v) in [bool, str, int, float, dict, list, tuple]
            },
            open(os.path.join(model_path, "training_args.json"), 'w'),
            ensure_ascii=False,
            indent=2)

        torch.save(optimizer.state_dict(),
                   os.path.join(model_path, "optimizer.pt"))
        torch.save(scheduler.state_dict(),
                   os.path.join(model_path, "scheduler.pt"))

    def train(self, args, train_examples, eval_examples):

        logger.info(
            f"Start train: {len(train_examples)} train examples, {len(eval_examples)} eval examples."
        )

        #  if is_master_process(args):
        #      tb_writer = SummaryWriter()

        train_dataset, _ = self.examples_to_dataset(train_examples,
                                                    args.train_max_seq_length)
        train_dataloader = generate_dataloader(
            args,
            train_dataset,
            batch_size=args.per_gpu_train_batch_size,
            keep_order=False,
            collate_fn=self.collate_fn)
        logger.info(f"Start training ...")
        logger.info(f"  Num examples    = {len(train_examples)}")
        logger.info(f"  Num epoch steps = {len(train_dataloader)}")
        logger.info(f"  Num epochs = {args.num_train_epochs}")
        logger.info(f"  Batch size = {args.per_gpu_train_batch_size}")

        steps_per_epoch = len(
            train_dataloader) // args.gradient_accumulation_steps

        if args.max_steps > 0:
            total_steps = args.max_steps
            args.num_train_epochs = args.max_steps // steps_per_epoch + 1
        else:
            total_steps = steps_per_epoch * args.num_train_epochs
        args.total_steps = total_steps

        logger.info(
            f"  Gradient Accumulation steps = {args.gradient_accumulation_steps}"
        )
        logger.info(f"  Total optimization steps = {total_steps}")

        model, optimizer, scheduler = self.build_model(args)
        tokenizer = self.tokenizer

        # Check if saved optimizer or scheduler states exist
        model_path = Path(args.model_path)
        optimizer_saved_file = model_path / "optimizer.pt"
        scheduler_saved_file = model_path / "scheduler.pt"
        if optimizer_saved_file.exists() and scheduler_saved_file.exists():
            optimizer.load_state_dict(torch.load(optimizer_saved_file))
            scheduler.load_state_dict(torch.load(scheduler_saved_file))

        if args.fp16:
            try:
                from apex import amp
            except ImportError:
                raise ImportError(
                    "Please install apex from https://github.com/nvidia/apex to use fp16."
                )
            model, optimizer = amp.initialize(model,
                                              optimizer,
                                              opt_level=args.fp16_opt_level)
            args.amp = amp
        # multi-gpu training (should be after apex fp16 initialization)
        if args.n_gpu > 1:
            model = torch.nn.DataParallel(model)
        # Distributed training (should be after apex fp16 initialization)
        #  if args.local_rank != -1:
        if is_multi_processes(args):
            model = torch.nn.parallel.DistributedDataParallel(
                model,
                device_ids=[args.local_rank],
                output_device=args.local_rank,
                find_unused_parameters=True)

        trained_steps = 0
        epochs_trained = 0
        steps_trained_in_current_epoch = 0
        # Check if continuing training from a checkpoint
        model_path = Path(args.model_path)
        output_dir = Path(args.output_dir)

        if model_path.exists() and "checkpoint" in str(model_path):
            # set trained_steps to trained_steps of last saved checkpoint from model path
            trained_steps = int(model_path.parts()[-1].split("-")[-1])
            epochs_trained = trained_steps // (
                len(train_dataloader) // args.gradient_accumulation_steps)
            steps_trained_in_current_epoch = trained_steps % (
                len(train_dataloader) // args.gradient_accumulation_steps)

            logger.info(
                "  Continuing training from checkpoint, will skip to saved trained_steps"
            )
            logger.info("  Continuing training from epoch %d", epochs_trained)
            logger.info("  Continuing training from global step %d",
                        trained_steps)
            logger.info("  Will skip the first %d steps in the first epoch",
                        steps_trained_in_current_epoch)

        train_loss, logging_loss = 0.0, 0.0
        best_value = 0.0
        best_type = 'max'  # ['max', 'min']

        #  best_value = float('inf')
        #  best_type = 'min'

        model.zero_grad()

        # https://arxiv.org/abs/2002.10345
        # https://github.com/lonePatient/BERT-SDA
        enable_kd = args.enable_kd
        if enable_kd:
            logger.warning("Enable knowledge distillation.")
            from torch.nn import MSELoss
            kd_loss_fct = MSELoss()
            kd_model = copy.deepcopy(model)
            kd_model.eval()

        #  train_iterator = trange(
        #      epochs_trained,
        #      int(args.num_train_epochs),
        #      desc="Epoch",
        #      disable=args.local_rank not in [-1, 0],
        #  )
        #  for epoch in train_iterator:
        for epoch in range(epochs_trained, args.num_train_epochs):

            #  epoch_iterator = tqdm(train_dataloader,
            #                        desc="Iteration",
            #                        disable=args.local_rank not in [-1, 0])
            #  for step, batch in enumerate(epoch_iterator):

            pbar = Progbar(target=len(train_dataloader),
                           stateful_metrics=['loss'],
                           desc=f"Epoch({epoch+1}/{args.num_train_epochs})")
            for step, batch in enumerate(train_dataloader):
                if steps_trained_in_current_epoch > 0:
                    steps_trained_in_current_epoch -= 1
                    continue

                model.train()
                batch = tuple(t.to(args.device) for t in batch)

                outputs = self.on_train_step(args, model, step, batch)
                #  inputs = self.batch_to_inputs(args, batch)
                #  outputs = model(**inputs)

                #  logger.debug(f"outputs: {outputs}")
                # -------- loss --------
                loss = outputs[0]

                if enable_kd:
                    inputs = self.batch_to_inputs(args, batch)
                    if "labels" in inputs:
                        inputs['labels'] = None
                    #  num_inputs = len(inputs)
                    #  for kk, vv in inputs.items():
                    #      logger.info(f"inputs[{kk}]: {vv.shape}, {vv}")
                    with torch.no_grad():
                        kd_logits = kd_model(**inputs)[0]
                        #  kd_logits = kd_model(**inputs)[1]
                    #  logger.warning(
                    #      f"outputs[1]: {outputs[1].shape}, {outputs[1]}")
                    #  logger.warning(
                    #      f"outputs[0]: {outputs[0].shape}, {outputs[0]}")
                    #  logger.warning(
                    #      f"kd_logits: {kd_logits.shape}, {kd_logits}")
                    kd_loss = kd_loss_fct(outputs[1], kd_logits)
                    loss += args.kd_coeff * kd_loss

                #  logger.debug(f"loss: {loss}")
                if args.n_gpu > 1:
                    loss = loss.mean()
                if args.gradient_accumulation_steps > 1:
                    loss = loss / args.gradient_accumulation_steps

                if args.fp16:
                    with amp.scale_loss(loss, optimizer) as scaled_loss:
                        scaled_loss.backward()
                else:
                    loss.backward()

                lr = scheduler.get_last_lr()[0]
                #  lr = scheduler.get_lr()[0]
                pbar.update(step + 1,
                            values=[('lr', lr), ('loss', loss.item())])
                train_loss += loss.item()

                if (step + 1) % args.gradient_accumulation_steps == 0:
                    if args.fp16:
                        torch.nn.utils.clip_grad_norm_(
                            amp.master_params(optimizer), args.max_grad_norm)
                    else:
                        torch.nn.utils.clip_grad_norm_(model.parameters(),
                                                       args.max_grad_norm)
                    optimizer.step()
                    scheduler.step()
                    model.zero_grad()
                    trained_steps += 1

                    if enable_kd:
                        decay = min(args.kd_decay,
                                    (1 + trained_steps) / (10 + trained_steps))
                        one_minus_decay = 1.0 - decay
                        with torch.no_grad():
                            parameters = [
                                p for p in model.parameters()
                                if p.requires_grad
                            ]
                            for s_param, param in zip(kd_model.parameters(),
                                                      parameters):
                                s_param.sub_(one_minus_decay *
                                             (s_param - param))

                # -------- Save models --------
                if is_master_process(
                        args) and trained_steps % steps_per_epoch == 0:

                    # -------- Save checkpoint --------
                    if args.save_checkpoints:
                        checkpoint_dir = f"checkpoint-{trained_steps}"
                        checkpoint_path = output_dir / checkpoint_dir
                        self.save_model(args, model, tokenizer, optimizer,
                                        scheduler, checkpoint_path)

                    # -------- Evaluate --------
                    if not args.no_eval_on_each_epoch:
                        logger.info(
                            f"Epoch({epoch+1}/{args.num_train_epochs}) evaluating."
                        )
                        eval_logs = {}
                        eval_results = self.evaluate(args, model,
                                                     eval_examples)
                        for key, value in eval_results.items():
                            eval_key = "eval_{}".format(key)
                            eval_logs[eval_key] = f"{value:.6f}"
                        loss_scalar = (train_loss -
                                       logging_loss) / steps_per_epoch
                        learning_rate_scalar = scheduler.get_last_lr()[0]
                        #  learning_rate_scalar = scheduler.get_lr()[0]
                        eval_logs[
                            "learning_rate"] = f"{learning_rate_scalar:.6f}"
                        eval_logs["loss"] = f"{loss_scalar:.6f}"
                        #  for key, value in eval_logs.items():
                        #      tb_writer.add_scalar(key, value, trained_steps)
                        print(
                            json.dumps({
                                **eval_logs,
                                **{
                                    "step": trained_steps
                                }
                            }))
                        logging_loss = train_loss

                        # -------- Save best model --------
                        best_index = 'f1'  # ['f1', 'acc', 'recall', 'loss']
                        eval_value = eval_results[best_index]
                        is_best = False
                        if best_index in ['loss']:
                            if eval_value < best_value:
                                is_best = True
                        else:
                            if eval_value > best_value:
                                is_best = True
                        if is_best:
                            logger.info(
                                f"Best {best_index}: {eval_value:.4f} ({eval_value - best_value:.6f})"
                            )
                            best_value = eval_value
                            bestmodel_path = output_dir / f"best_fold{args.fold}"
                            self.save_model(args, model, tokenizer, optimizer,
                                            scheduler, bestmodel_path)

                            if args.save_checkpoints:
                                best_symlink = output_dir / "best/best_checkpoint"
                                if best_symlink.is_symlink():
                                    best_symlink.unlink()
                                os.symlink(f"../{checkpoint_dir}",
                                           best_symlink)

            print(" ")
            if 'cuda' in str(args.device):
                torch.cuda.empty_cache()

        #      if args.max_steps > 0 and trained_steps > args.max_steps:
        #          epoch_iterator.close()
        #          break
        #  if args.max_steps > 0 and trained_steps > args.max_steps:
        #      train_iterator.close()
        #      break

        #  if is_multi_processes(args):
        #      tb_writer.close()
        return trained_steps, train_loss / trained_steps

    def evaluate(self, args, model, eval_examples):
        eval_dataset, eval_features = self.examples_to_dataset(
            eval_examples, args.eval_max_seq_length)

        self.on_eval_start(args, eval_dataset)
        eval_dataloader = generate_dataloader(
            args,
            eval_dataset,
            batch_size=args.per_gpu_eval_batch_size,
            keep_order=True,
            collate_fn=self.collate_fn)

        eval_output_dir = args.output_dir
        if is_multi_processes(args) and not os.path.exists(eval_output_dir):
            os.makedirs(eval_output_dir)

        logger.info(f"Start evaluating ...")
        logger.info(f"  Num examples    = {len(eval_examples)}")
        logger.info(f"  Num epoch steps = {len(eval_dataloader)}")
        logger.info(f"  Batch size = {args.per_gpu_eval_batch_size}")

        eval_steps = 0
        eval_loss = 0.0
        pbar = Progbar(target=len(eval_dataloader),
                       stateful_metrics=['loss', 'acc', 'recall', 'f1'],
                       desc=f"Evaluating")
        for step, batch in enumerate(eval_dataloader):
            model.eval()
            batch = tuple(t.to(args.device) for t in batch)
            with torch.no_grad():
                #  inputs = self.batch_to_inputs(args, batch)
                #  outputs = model(**inputs)
                batch_features = eval_features[args.per_gpu_eval_batch_size *
                                               step:args.
                                               per_gpu_eval_batch_size *
                                               (step + 1)]
                outputs, results = self.on_eval_step(args, model, step, batch,
                                                     batch_features)

                # -------- loss --------
                loss = outputs[0]

                if args.n_gpu > 1:
                    loss = loss.mean()
                eval_loss += loss.item()

            eval_steps += 1

            #  results = self.on_eval_step(args, eval_dataset, step, model,
            #                              inputs, outputs)
            values = []
            if results:
                values = [(k, v) for k, v in results.items()]
            values += [('loss', loss.item())]
            pbar.update(step + 1, values=values)

        self.eval_loss = eval_loss / eval_steps
        results = self.on_eval_end(args, eval_dataset)

        return results

    def predict(self, args, model, test_examples):

        pred_output_dir = Path(args.output_dir)
        if is_multi_processes(args) and not pred_output_dir.exists():
            os.makedirs(pred_output_dir)

        test_dataset, _ = self.examples_to_dataset(test_examples,
                                                   args.eval_max_seq_length)
        test_dataloader = generate_dataloader(
            args,
            test_dataset,
            batch_size=args.per_gpu_predict_batch_size,
            keep_order=True,
            collate_fn=self.collate_fn)

        logger.info(f"Start predicting ...")
        logger.info(f"  Num examples    = {len(test_examples)}")
        logger.info(f"  Num epoch steps = {len(test_dataloader)}")
        logger.info(f"  Batch size = {args.per_gpu_predict_batch_size}")

        pbar = Progbar(target=len(test_dataloader), desc=f"Predicting")
        self.on_predict_start(args, test_dataset)
        for step, batch in enumerate(test_dataloader):
            model.eval()
            batch = tuple(t.to(args.device) for t in batch)
            with torch.no_grad():
                #  inputs = self.package_inputs_from_batch(args,
                #                                          batch,
                #                                          known_labels=False)
                #  outputs = model(**inputs)
                #  self.on_predict_step(args, test_dataset, step, model, inputs,
                #                       outputs)
                self.on_predict_step(args, model, step, batch)

            pbar.update(step + 1)

        results = self.on_predict_end(args, test_dataset)
        return results



import torch
import torch.nn as nn
import torch.nn.functional as F

class FeedForwardNetwork(nn.Module):
    def __init__(self, input_size, hidden_size, output_size, dropout_rate=0):
        super(FeedForwardNetwork, self).__init__()
        self.dropout_rate = dropout_rate
        self.linear1 = nn.Linear(input_size, hidden_size)
        self.linear2 = nn.Linear(hidden_size, output_size)

    def forward(self, x):
        x_proj = F.dropout(F.relu(self.linear1(x)), p=self.dropout_rate, training=self.training)
        x_proj = self.linear2(x_proj)
        return x_proj


class PoolerStartLogits(nn.Module):
    def __init__(self, hidden_size, num_classes):
        super(PoolerStartLogits, self).__init__()
        self.dense = nn.Linear(hidden_size, num_classes)

    def forward(self, hidden_states, p_mask=None):
        x = self.dense(hidden_states)
        return x

class PoolerEndLogits(nn.Module):
    def __init__(self, hidden_size, num_classes):
        super(PoolerEndLogits, self).__init__()
        self.dense_0 = nn.Linear(hidden_size, hidden_size)
        self.activation = nn.Tanh()
        self.LayerNorm = nn.LayerNorm(hidden_size)
        self.dense_1 = nn.Linear(hidden_size, num_classes)

    def forward(self, hidden_states, start_positions=None, p_mask=None):
        x = self.dense_0(torch.cat([hidden_states, start_positions], dim=-1))
        x = self.activation(x)
        x = self.LayerNorm(x)
        x = self.dense_1(x)
        return x



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os, sys, time
from pathlib import Path
from loguru import logger
from .trainer import generate_dataloader
from .onnx import export_onnx, inference_from_onnx
from .ner_utils import LabeledText, show_ner_datainfo, get_ner_preds_reviews, save_ner_preds, load_ner_examples, load_ner_labeled_examples
from .glue_utils import show_glue_datainfo, load_glue_examples
from .common_args import add_common_args


def augement_entities(all_text_entities, labels_map):
    aug_tokens = []
    for i, (guid, text, entities) in enumerate(
            tqdm(all_text_entities, desc=f"Augement {num_augements}X")):

        #  print(f"-------------------{json_file}--------------------")
        #  print(text)
        #  print(entities)
        #  for entity in entities:
        #      s = entity['start_pos']
        #      e = entity['end_pos']
        #      print(f"{entity['label_type']}: {text[s:e]}")
        #  print("----------------------------------------")
        if entities:
            for ai in range(num_augements):
                e_idx = random.randint(0, len(entities) - 1)
                entity = entities[e_idx]

                label_type = entity['label_type']
                s = entity['start_pos']
                e = entity['end_pos']

                labels = labels_map[label_type]
                idx = random.randint(0, len(labels) - 1)
                new_entity_text = labels[idx]

                text = text[:s] + new_entity_text + text[e:]

                assert len(new_entity_text) >= 0
                delta = len(new_entity_text) - (e - s)

                entity['end_pos'] = entity['start_pos'] + len(new_entity_text)
                entity['mention'] = new_entity_text

                assert text[
                    entity['start_pos']:entity['end_pos']] == new_entity_text

                for n, e in enumerate(entities):
                    if n > e_idx:
                        e['start_pos'] += delta
                        e['end_pos'] += delta

                aug_tokens.append(
                    (f"{guid}-a{ai}", text, copy.deepcopy(entities)))

    #  for guid, text, entities in aug_tokens:
    #      text_a = text
    #      for entity in entities:
    #          logger.debug(f"{guid}: text_a: {text_a}")
    #          logger.debug(
    #              f"text_a[entity['start_pos']:entity['end_pos']]: {text_a[entity['start_pos']:entity['end_pos']]}"
    #          )
    #          logger.debug(
    #              f"mention {entity['mention']} in {text_a.find(entity['mention'])}"
    #          )
    #          logger.debug(f"entity: {entity}")
    #          assert text_a[entity['start_pos']:entity['end_pos']] == entity[
    #              'mention']

    return aug_tokens


#  def data_seg_generator(lines, ner_labels, seg_len=0, seg_backoff=0):
#      all_text_entities = []
#      labels_map = {}
#
#      for i, s in enumerate(tqdm(lines)):
#          guid = str(i)
#          text = s['originalText'].strip()
#          entities = s['entities']
#
#          new_entities = []
#          used_span = []
#          entities = sorted(entities, key=lambda e: e['start_pos'])
#          for entity in entities:
#              if entity['label_type'] not in ner_labels:
#                  continue
#              entity['mention'] = text[entity['start_pos']:entity['end_pos']]
#              s = entity['start_pos']
#              e = entity['end_pos']
#
#              overlap = False
#              for us in used_span:
#                  if s >= us[0] and s < us[1]:
#                      overlap = True
#                      break
#                  if e > us[0] and e <= us[1]:
#                      overlap = True
#                      break
#              if overlap:
#                  logger.warning(
#                      f"Overlap! {i} mention: {entity['mention']}, used_span: {used_span}"
#                  )
#                  continue
#              used_span.append((s, e))
#
#              new_entities.append(entity)
#          entities = new_entities
#
#          guid = str(i)
#
#          seg_offset = 0
#          if seg_len <= 0:
#              seg_len = max_seq_length
#
#          for (seg_text, ) in seg_generator((text, ), seg_len, seg_backoff):
#              text_a = seg_text
#
#              seg_start = seg_offset
#              seg_end = seg_offset + min(seg_len, len(seg_text))
#              labels = [
#                  (x['label_type'], x['start_pos'] - seg_offset,
#                   x['end_pos'] - 1 - seg_offset) for x in entities
#                  if x['start_pos'] >= seg_offset and x['end_pos'] <= seg_end
#              ]
#
#              # 没有标注存在的文本片断不用于训练
#              if labels:
#                  yield guid, text_a, None, labels
#
#                  if num_augements > 0:
#                      seg_entities = [{
#                          'start_pos': x['start_pos'] - seg_offset,
#                          'end_pos': x['end_pos'] - seg_offset,
#                          'label_type': x['label_type'],
#                          'mention': x['mention']
#                      } for x in entities if x['start_pos'] >= seg_offset
#                                      and x['end_pos'] <= seg_end]
#                      all_text_entities.append((guid, text_a, seg_entities))
#
#                      for entity in seg_entities:
#                          label_type = entity['label_type']
#                          s = entity['start_pos']  # - seg_offset
#                          e = entity['end_pos']  #- seg_offset
#                          #  print(s, e)
#                          assert e >= s
#                          #  logger.debug(
#                          #      f"seg_start: {seg_start}, seg_end: {seg_end}, seg_offset: {seg_offset}"
#                          #  )
#                          #  logger.debug(f"s: {s}, e: {e}")
#                          assert s >= 0 and e <= len(seg_text)
#                          #  if s >= len(seg_text) or e >= len(seg_text):
#                          #      continue
#
#                          entity_text = seg_text[s:e]
#                          #  print(label_type, entity_text)
#
#                          assert len(entity_text) > 0
#                          if label_type not in labels_map:
#                              labels_map[label_type] = []
#                          labels_map[label_type].append(entity_text)
#
#              seg_offset += seg_len - seg_backoff
#
#      if num_augements > 0:
#          aug_tokens = augement_entities(all_text_entities, labels_map)
#          for guid, text, entities in aug_tokens:
#              text_a = text
#              for entity in entities:
#                  #  logger.debug(f"text_a: {text_a}")
#                  #  logger.debug(
#                  #      f"text_a[entity['start_pos']:entity['end_pos']]: {text_a[entity['start_pos']:entity['end_pos']]}"
#                  #  )
#                  #  logger.debug(
#                  #      f"mention {entity['mention']} in {text_a.find(entity['mention'])}"
#                  #  )
#                  #  logger.debug(f"entity: {entity}")
#                  assert text_a[entity['start_pos']:entity['end_pos']] == entity[
#                      'mention']
#              labels = [
#                  (entity['label_type'], entity['start_pos'],
#                   entity['end_pos'] - 1) for entity in entities
#                  if entity['end_pos'] <= (
#                      min(len(text_a), seg_len) if seg_len > 0 else len(text_a))
#              ]
#              yield guid, text_a, None, labels



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import random, copy, json
from tqdm import tqdm
from loguru import logger
from theta.utils import seg_generator
from dataclasses import dataclass, field
from typing import List

# ----------------------
# 实体标注规范
#  {
#      'guid':
#      "0",
#      'text':
#      'xxxxxxx',
#      'entities': [{
#          'category': category,
#          'start': start,
#          'end': end,
#          'mention': mention
#      }]
#  }


class DataClassBase:
    def to_dict(self):
        #  return self.__dict__
        dict_obj = {}
        for k, v in self.__dict__.items():
            if k.startswith('_'):
                continue
            if isinstance(v, List):
                dict_obj[k] = [
                    x.to_dict() if isinstance(x, DataClassBase) else x
                    for x in v
                ]
            else:
                dict_obj[k] = v
        return dict_obj

    def from_dict(self, dict_data):
        for k in self.__dict__.keys():
            if k in dict_data:
                setattr(self, k, dict_data[k])


@dataclass(unsafe_hash=True)
class Entity(DataClassBase):
    category: str = field(default_factory=str)
    start: int = field(default_factory=int)
    end: int = field(default_factory=int)
    mention: str = field(default_factory=str)


@dataclass
class LabeledText(DataClassBase):
    guid: str = field(default_factory=str)
    text: str = field(default_factory=str)
    entities: List[Entity] = field(default_factory=list)

    def add_entity(self, category, start, end):
        self.entities.append(
            Entity(category=category,
                   start=start,
                   end=end,
                   mention=self.text[start:end + 1]))

    def offset(self, text_offset):
        for entity in self.entities:
            entity.start += text_offset
            entity.end += text_offset

    def get_annotated_text(self, seg_len, seg_backoff):
        #  debug = False
        #  if self.guid == '449':
        #      debug = True

        max_text_length = seg_len - seg_backoff

        text = self.text[:max_text_length]
        #  if debug:
        #      logger.debug(f"text: <{text}>")
        annotated_text = ""
        p0 = 0
        for entity in self.entities:
            x0 = entity.start
            x1 = entity.end
            if x1 >= max_text_length:
                continue
            annotated_text += text[p0:x0]
            #  if debug:
            #      logger.debug(
            #          f"annotated_text(p0:{p0},x0:{x0},x1:{x1}): {annotated_text}"
            #      )
            annotated_text += f"【{entity.mention}】"
            #  if debug:
            #      logger.debug(
            #          f"annotated_text(p0:{p0},x0:{x0},x1:{x1}): {annotated_text}"
            #      )
            p0 = x1 + 1
        annotated_text += text[p0:]
        #  if debug:
        #      logger.debug(f"annotated_text(p0:{p0}): {annotated_text}")

        return annotated_text

    #  def to_json(self):
    #      json_entities = [x.to_dict() for x in self.entities]
    #      json_data = {
    #          'guid': self.guid,
    #          'text': self.text,
    #          'entities': json_entities
    #      }
    #      return json_data

    #  def from_json(self, json_data):
    #      self.guid = json_data['guid']
    #      self.text = json_data['text']
    #      self.entities = []
    #      for x in json_data['entities']:
    #          entity = Entity()
    #          entity.from_dict(x)
    #          self.entities.append(entity)


def get_ner_preds_reviews(preds, examples, seg_len, seg_backoff):
    reviews = {}
    category_mentions = {}
    for json_d, input_example in tqdm(zip(preds, examples)):
        guid = input_example.guid
        text_offset = input_example.text_offset
        text = input_example.text_a
        entities = json_d['entities']

        labeled_text = LabeledText(guid, text)
        for c, x0, x1 in entities:
            labeled_text.add_entity(c, x0, x1)
            if c not in category_mentions:
                category_mentions[c] = set()
            category_mentions[c].add(labeled_text.entities[-1].mention)

        annotated_text = labeled_text.get_annotated_text(seg_len, seg_backoff)

        labeled_text.offset(text_offset)
        json_data = labeled_text.to_dict()

        if guid not in reviews:
            reviews[guid] = {
                'guid': guid,
                'text': "",
                'annotated_text': "",
                'entities': []
            }
        reviews[guid]['text'] += text[:seg_len - seg_backoff]
        reviews[guid]['annotated_text'] += annotated_text
        for x in json_data['entities']:
            if x not in reviews[guid]['entities']:
                reviews[guid]['entities'].append(x)
        reviews[guid]['entities'] = sorted(reviews[guid]['entities'],
                                           key=lambda x: x['start'])

    return reviews, category_mentions


def save_ner_preds(args, preds, test_examples):
    reviews, category_mentions = get_ner_preds_reviews(preds, test_examples,
                                                       args.seg_len,
                                                       args.seg_backoff)

    reviews_file = f"{args.output_dir}/{args.dataset_name}_reviews_fold{args.fold}.json"
    json.dump(reviews, open(reviews_file, 'w'), ensure_ascii=False, indent=2)
    logger.info(f"Reviews file: {reviews_file}")

    category_mentions_file = f"{args.output_dir}/{args.dataset_name}_category_mentions_fold{args.fold}.txt"
    num_categories = len(category_mentions)
    num_mentions = 0
    with open(category_mentions_file, 'w') as wt:
        for c, mentions in category_mentions.items():
            for m in mentions:
                wt.write(f"{c}\t{m}\n")
                num_mentions += 1
    logger.info(
        f"Total {num_categories} categories and {num_mentions} mentions saved to {category_mentions_file}"
    )


def augement_entities(all_text_entities, labels_map, num_augements):
    aug_tokens = []
    for i, (guid, text, entities) in enumerate(
            tqdm(all_text_entities, desc=f"Augement {num_augements}X")):

        #  print(f"-------------------{json_file}--------------------")
        #  print(text)
        #  print(entities)
        #  for entity in entities:
        #      s = entity['start_pos']
        #      e = entity['end_pos']
        #      print(f"{entity['label_type']}: {text[s:e]}")
        #  print("----------------------------------------")
        if entities:
            for ai in range(num_augements):
                e_idx = random.randint(0, len(entities) - 1)
                entity = entities[e_idx]

                label_type = entity['category']
                s = entity['start']
                e = entity['end']

                labels = labels_map[label_type]
                idx = random.randint(0, len(labels) - 1)
                new_entity_text = labels[idx]

                text = text[:s] + new_entity_text + text[e + 1:]

                assert len(new_entity_text) >= 0
                delta = len(new_entity_text) - (e + 1 - s)

                entity['end'] = entity['start'] + len(new_entity_text) - 1
                entity['mention'] = new_entity_text

                assert text[entity['start']:entity['end'] +
                            1] == new_entity_text

                for n, e in enumerate(entities):
                    if n > e_idx:
                        e['start'] += delta
                        e['end'] += delta

                aug_tokens.append(
                    (f"{guid}-a{ai}", text, copy.deepcopy(entities)))

    #  for guid, text, entities in aug_tokens:
    #      text_a = text
    #      for entity in entities:
    #          logger.debug(f"{guid}: text_a: {text_a}")
    #          logger.debug(
    #              f"text_a[entity['start']:entity['end']]: {text_a[entity['start']:entity['end']]}"
    #          )
    #          logger.debug(
    #              f"mention {entity['mention']} in {text_a.find(entity['mention'])}"
    #          )
    #          logger.debug(f"entity: {entity}")
    #          assert text_a[entity['start']:entity['end']] == entity[
    #              'mention']

    return aug_tokens


def data_seg_generator(lines,
                       ner_labels,
                       seg_len=0,
                       seg_backoff=0,
                       num_augements=0,
                       allow_overlap=False):
    all_text_entities = []
    labels_map = {}

    for i, s in enumerate(tqdm(lines)):
        #  logger.debug(f"s: {s}")
        guid = str(i)
        text = s['text'].strip()
        entities = s['entities']

        new_entities = []
        used_span = []
        #  logger.debug(f"entities: {entities}")
        entities = sorted(entities, key=lambda e: e.start)
        for entity in entities:
            if entity.category not in ner_labels:
                continue
            entity.mention = text[entity.start:entity.end + 1]
            s = entity.start
            e = entity.end

            overlap = False
            for us in used_span:
                if s >= us[0] and s <= us[1]:
                    overlap = True
                    break
                if e >= us[0] and e <= us[1]:
                    overlap = True
                    break
            if overlap and not allow_overlap:
                logger.warning(
                    f"Overlap! {i} mention: {entity.mention}({s}:{e}), used_span: {used_span}"
                )
                continue
            used_span.append((s, e))

            new_entities.append(entity)
        entities = new_entities

        seg_offset = 0
        if seg_len <= 0:
            seg_len = max_seq_length

        for (seg_text, ), text_offset in seg_generator((text, ), seg_len,
                                                       seg_backoff):
            text_a = seg_text

            assert text_offset == seg_offset

            seg_start = seg_offset
            seg_end = seg_offset + min(seg_len, len(seg_text))
            labels = [(x.category, x.start - seg_offset, x.end - seg_offset)
                      for x in entities
                      if x.start >= seg_offset and x.end < seg_end]

            # 没有标注存在的文本片断不用于训练
            if labels:
                yield guid, text_a, None, labels, seg_offset

                if num_augements > 0:
                    seg_entities = [
                        {
                            'start': x.start - seg_offset,
                            'end': x.end - seg_offset,
                            'category': x.category,
                            'mention': x.mention
                        } for x in entities
                        if x.start >= seg_offset and x.end < seg_end
                    ]
                    all_text_entities.append((guid, text_a, seg_entities))

                    for entity in seg_entities:
                        label_type = entity['category']
                        s = entity['start']  # - seg_offset
                        e = entity['end']  #- seg_offset
                        #  print(s, e)
                        assert e >= s
                        #  logger.debug(
                        #      f"seg_start: {seg_start}, seg_end: {seg_end}, seg_offset: {seg_offset}"
                        #  )
                        #  logger.debug(f"s: {s}, e: {e}")
                        assert s >= 0 and e < len(seg_text)
                        #  if s >= len(seg_text) or e >= len(seg_text):
                        #      continue

                        entity_text = seg_text[s:e + 1]
                        #  print(label_type, entity_text)

                        assert len(entity_text) > 0
                        if label_type not in labels_map:
                            labels_map[label_type] = []
                        labels_map[label_type].append(entity_text)

            seg_offset += seg_len - seg_backoff

    if num_augements > 0:
        aug_tokens = augement_entities(all_text_entities,
                                       labels_map,
                                       num_augements=num_augements)
        for guid, text, entities in aug_tokens:
            text_a = text
            for entity in entities:
                #  logger.debug(f"text_a: {text_a}")
                #  logger.debug(
                #      f"text_a[entity['start']:entity['end']]: {text_a[entity['start']:entity['end']]}"
                #  )
                #  logger.debug(
                #      f"mention {entity['mention']} in {text_a.find(entity['mention'])}"
                #  )
                #  logger.debug(f"entity: {entity}")
                assert text_a[entity['start']:entity['end'] +
                              1] == entity['mention']
            labels = [
                (entity['category'], entity['start'], entity['end'])
                for entity in entities if entity['end'] < (
                    min(len(text_a), seg_len) if seg_len > 0 else len(text_a))
            ]
            yield guid, text_a, None, labels, 0


class InputExample(object):
    """A single training/test example for token classification."""
    def __init__(self, guid, text_a, labels, text_offset=0):
        self.guid = guid
        self.text_a = text_a
        self.labels = labels
        self.text_offset = text_offset

    def __repr__(self):
        return str(self.to_json_string())

    def to_dict(self):
        """Serializes this instance to a Python dictionary."""
        output = copy.deepcopy(self.__dict__)
        return output

    def to_json_string(self):
        """Serializes this instance to a JSON string."""
        return json.dumps(self.to_dict(), indent=2, sort_keys=True) + "\n"


def load_ner_examples(data_generator, examples_file, seg_len=0, seg_backoff=0):

    examples = []

    for guid, text_a, _, labels in data_generator(examples_file):
        assert text_a is not None
        for (seg_text_a, ), text_offset in seg_generator((text_a, ), seg_len,
                                                         seg_backoff):
            examples.append(
                InputExample(guid=guid,
                             text_a=seg_text_a,
                             labels=labels,
                             text_offset=text_offset))
    logger.info(f"Loaded {len(examples)} examples.")

    return examples


def load_ner_labeled_examples(lines,
                              ner_labels,
                              seg_len=0,
                              seg_backoff=0,
                              num_augements=0,
                              allow_overlap=False):

    examples = []
    for guid, text_a, text_b, labels, text_offset in data_seg_generator(
            lines,
            ner_labels,
            seg_len=seg_len,
            seg_backoff=seg_backoff,
            num_augements=num_augements,
            allow_overlap=allow_overlap):
        assert text_a is not None
        examples.append(
            InputExample(guid=guid,
                         text_a=text_a,
                         labels=labels,
                         text_offset=text_offset))

    logger.info(f"Loaded {len(examples)} examples.")
    return examples


def show_ner_datainfo(ner_labels, train_data_generator, train_file,
                      test_data_generator, test_file):
    train_lengths = []
    for guid, text_a, _, labels in train_data_generator(train_file):
        train_lengths.append(len(text_a))
        for entity in labels:
            c = entity.category
            if c not in ner_labels:
                ner_labels.append(c)

    test_lengths = [
        len(text_a)
        for guid, text_a, _, labels in test_data_generator(test_file)
    ]

    logger.info(f"****** ner_labels ******")
    logger.info(f"{ner_labels}")
    logger.info(f"")
    import numpy as np
    logger.info(f"****** train lengths ******")
    logger.info(f"mean: {np.mean(train_lengths):.2f}")
    logger.info(f"std: {np.std(train_lengths):.2f}")
    logger.info(f"min: {np.min(train_lengths)}")
    logger.info(f"max: {np.max(train_lengths)}")
    logger.info(f"")

    logger.info(f"****** test lengths ******")
    logger.info(f"mean: {np.mean(test_lengths):.2f}")
    logger.info(f"std: {np.std(test_lengths):.2f}")
    logger.info(f"min: {np.min(test_lengths)}")
    logger.info(f"max: {np.max(test_lengths)}")
    logger.info(f"")
